"""
Managers for our models
"""

from django.db import models
from django.db.models import Count, F


class AFatLinkQuerySet(models.QuerySet):
    """
    AFatLinkQuerySet
    """

    def annotate_afats_count(self):
        """
        Annotate the amount fats per fat link
        :return:
        :rtype:
        """

        return self.annotate(afats_count=Count(F("afats")))


class AFatLinkManager(models.Manager):
    """
    AFatLinkManager
    """

    def get_queryset(self) -> models.QuerySet:
        """
        Integrate custom QuerySet methods.
        """

        return AFatLinkQuerySet(self.model, using=self._db)

    def select_related_default(self):
        """
        Apply select_related for default query optimizations.
        """

        return self.select_related(
            "link_type", "creator", "character", "creator__profile__main_character"
        )


class AFatManager(models.Manager):
    """
    AFatManager
    """

    def select_related_default(self):
        """
        Apply select_related for default query optimizations.
        """

        return self.select_related("afatlink", "afatlink__link_type", "character")



# Generated by Django 3.1.8 on 2021-04-20 20:16

import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import afat.models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0014_auto_20210105_1413"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("afat", "0017_remove_soft_deletion"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="aaafat",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access the AFAT module"),
                    ("manage_afat", "Can manage the AFAT module"),
                    ("add_fatlink", "Can create FAT Links"),
                    ("stats_corporation_own", "Can see own corporation statistics"),
                    (
                        "stats_corporation_other",
                        "Can see statistics of other corporations",
                    ),
                    ("log_view", "Can view the modules log"),
                ),
                "verbose_name": "Alliance Auth AFAT",
            },
        ),
        migrations.AlterModelOptions(
            name="manualafat",
            options={
                "default_permissions": (),
                "verbose_name": "Manual FAT",
                "verbose_name_plural": "Manual FATs",
            },
        ),
        migrations.AddField(
            model_name="afatlink",
            name="reopened",
            field=models.BooleanField(
                default=False, help_text="Has this FAT link being re-opened?"
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="afatlink",
            field=models.ForeignKey(
                help_text="The fatlink the character registered at",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="character",
            field=models.ForeignKey(
                help_text="Character who registered this fat",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="character",
            field=models.ForeignKey(
                default=None,
                help_text="Character this fatlink has been created with",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="creator",
            field=models.ForeignKey(
                help_text="Who created the fatlink?",
                on_delete=models.SET(afat.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="link_type",
            field=models.ForeignKey(
                help_text="The fatlinks fleet type, if it's set",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="afat.afatlinktype",
            ),
        ),
        migrations.AlterField(
            model_name="manualafat",
            name="afatlink",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="manualafat",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="manualafat",
            name="creator",
            field=models.ForeignKey(
                on_delete=models.SET(afat.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="AFatLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("log_time", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "log_event",
                    models.CharField(
                        choices=[
                            ("CR_FAT_LINK", "FAT Link Created"),
                            ("CH_FAT_LINK", "FAT Link Changed"),
                            ("RM_FAT_LINK", "FAT Link Removed"),
                            ("RO_FAT_LINK", "FAT Link Re-Opened"),
                            ("RM_FAT", "FAT Removed"),
                            ("CR_FAT_MAN", "Manual FAT Added"),
                        ],
                        default="CR_FAT_LINK",
                        max_length=11,
                    ),
                ),
                ("log_text", models.TextField()),
                ("fatlink_hash", models.CharField(max_length=254)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=models.SET(afat.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "AFAT Log",
                "verbose_name_plural": "AFAT Logs",
                "default_permissions": (),
            },
        ),
    ]



"""
URL configuration
"""

from django.conf.urls import url

from afat.views import dashboard, fatlinks, logs, statistics

app_name: str = "afat"

urlpatterns = [
    # Dashboard
    url(r"^$", dashboard.overview, name="dashboard"),
    # Stats main page
    url(r"^statistics/$", statistics.overview, name="statistics_overview"),
    url(
        r"^statistics/(?P<year>[0-9]+)/$",
        statistics.overview,
        name="statistics_overview",
    ),
    # Stats corp
    url(
        r"^statistics/corporation/$",
        statistics.corporation,
        name="statistics_corporation",
    ),
    url(
        r"^statistics/corporation/(?P<corpid>[0-9]+)/$",
        statistics.corporation,
        name="statistics_corporation",
    ),
    url(
        r"^statistics/corporation/(?P<corpid>[0-9]+)/(?P<year>[0-9]+)/$",
        statistics.corporation,
        name="statistics_corporation",
    ),
    url(
        (
            r"^statistics/corporation/"
            r"(?P<corpid>[0-9]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/$"
        ),
        statistics.corporation,
        name="statistics_corporation",
    ),
    # Stats char
    url(r"^statistics/character/$", statistics.character, name="statistics_character"),
    url(
        r"^statistics/character/(?P<charid>[0-9]+)/$",
        statistics.character,
        name="statistics_character",
    ),
    url(
        (
            r"^statistics/character/"
            r"(?P<charid>[0-9]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/$"
        ),
        statistics.character,
        name="statistics_character",
    ),
    # Stats alliance
    url(r"^statistics/alliance/$", statistics.alliance, name="statistics_alliance"),
    url(
        r"^statistics/alliance/(?P<allianceid>[0-9]+)/$",
        statistics.alliance,
        name="statistics_alliance",
    ),
    url(
        r"^statistics/alliance/(?P<allianceid>[0-9]+)/(?P<year>[0-9]+)/$",
        statistics.alliance,
        name="statistics_alliance",
    ),
    url(
        (
            r"^statistics/alliance/"
            r"(?P<allianceid>[0-9]+)/(?P<year>[0-9]+)/(?P<month>[0-9]+)/$"
        ),
        statistics.alliance,
        name="statistics_alliance",
    ),
    # Fat links list actions
    url(r"^fatlinks/$", fatlinks.overview, name="fatlinks_overview"),
    url(r"^fatlinks/(?P<year>[0-9]+)/$", fatlinks.overview, name="fatlinks_overview"),
    # Fat link actions
    url(r"^fatlink/add/$", fatlinks.add_fatlink, name="fatlinks_add_fatlink"),
    url(
        r"^fatlinks/link/create/esi-fatlink/$",
        fatlinks.create_esi_fatlink,
        name="fatlinks_create_esi_fatlink",
    ),
    url(
        r"^fatlink/create/esi-fatlink/callback/(?P<fatlink_hash>[a-zA-Z0-9]+)/$",
        fatlinks.create_esi_fatlink_callback,
        name="fatlinks_create_esi_fatlink_callback",
    ),
    url(
        r"^fatlink/create/clickable-fatlink/$",
        fatlinks.create_clickable_fatlink,
        name="fatlinks_create_clickable_fatlink",
    ),
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/details/$",
        fatlinks.details_fatlink,
        name="fatlinks_details_fatlink",
    ),
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/delete/$",
        fatlinks.delete_fatlink,
        name="fatlinks_delete_fatlink",
    ),
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/stop-esi-tracking/$",
        fatlinks.close_esi_fatlink,
        name="fatlinks_close_esi_fatlink",
    ),
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/re-open/$",
        fatlinks.reopen_fatlink,
        name="fatlinks_reopen_fatlink",
    ),
    # Fat actions
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/register/$",
        fatlinks.add_fat,
        name="fatlinks_add_fat",
    ),
    url(
        r"^fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/fat/(?P<fat>[0-9]+)/delete/$",
        fatlinks.delete_fat,
        name="fatlinks_delete_fat",
    ),
    # Log actions
    url(r"^logs/$", logs.overview, name="logs_overview"),
    # Ajax calls :: Dashboard
    url(
        r"^ajax/dashboard/fatlinks/recent/$",
        dashboard.ajax_get_recent_fatlinks,
        name="dashboard_ajax_get_recent_fatlinks",
    ),
    url(
        r"^ajax/dashboard/fats/recent/character/(?P<charid>[0-9]+)/$",
        dashboard.ajax_recent_get_fats_by_character,
        name="dashboard_ajax_get_recent_fats_by_character",
    ),
    # Ajax calls :: Fat links
    url(
        r"^ajax/fatlinks/fatlinks/year/(?P<year>[0-9]+)/$",
        fatlinks.ajax_get_fatlinks_by_year,
        name="fatlinks_ajax_get_fatlinks_by_year",
    ),
    url(
        r"^ajax/fatlinks/fatlink/(?P<fatlink_hash>[a-zA-Z0-9]+)/fats/$",
        fatlinks.ajax_get_fats_by_fatlink,
        name="fatlinks_ajax_get_fats_by_fatlink",
    ),
    # Ajax calls :: Logs
    url(r"^ajax/logs/$", logs.ajax_get_logs, name="logs_ajax_get_logs"),
]



"""
Import FAT data from ImicusFAT module
"""

from django.apps import apps
from django.core.management.base import BaseCommand

from afat import __version__ as afat_version_installed
from afat.models import (
    AFat,
    AFatLink,
    AFatLinkType,
    AFatLog,
    AFatLogEvent,
    ClickAFatDuration,
)


def get_input(text) -> str:
    """
    Wrapped input to enable import
    :param text:
    :type text:
    :return:
    :rtype:
    """

    return input(text)


def imicusfat_installed() -> bool:
    """
    Check if imicusfat is installed
    :return:
    :rtype:
    """

    return apps.is_installed("imicusfat")


if imicusfat_installed():
    import sys

    import requests
    from imicusfat import __version__ as ifat_version_installed
    from imicusfat.models import (
        ClickIFatDuration,
        IFat,
        IFatLink,
        IFatLinkType,
        ManualIFat,
    )
    from packaging.specifiers import SpecifierSet
    from packaging.version import parse as version_parse


class Command(BaseCommand):
    """
    Initial import of FAT data from AA FAT module
    """

    help = "Imports FAT data from ImicusFAT module"

    def latest_version_available(self, package_name):
        """
        Checking for the latest available version of a package on Pypi
        :param package_name:
        :type package_name:
        :return:
        :rtype:
        """

        current_python_version = version_parse(
            f"{sys.version_info.major}.{sys.version_info.minor}"
            f".{sys.version_info.micro}"
        )

        current_version = version_parse(ifat_version_installed)
        current_is_prerelease = (
            str(current_version) == str(ifat_version_installed)
            and current_version.is_prerelease
        )

        result = requests.get(
            f"https://pypi.org/pypi/{package_name}/json", timeout=(5, 30)
        )

        latest = None

        if result.status_code == requests.codes.ok:
            pypi_info = result.json()

            for release, release_details in pypi_info["releases"].items():
                release_detail = (
                    release_details[-1] if len(release_details) > 0 else None
                )
                if not release_detail or (
                    not release_detail["yanked"]
                    and (
                        "requires_python" not in release_detail
                        or not release_detail["requires_python"]
                        or current_python_version
                        in SpecifierSet(release_detail["requires_python"])
                    )
                ):
                    my_release = version_parse(release)

                    if str(my_release) == str(release) and (
                        current_is_prerelease or not my_release.is_prerelease
                    ):
                        latest = release

            if not latest:
                self.stdout.write(
                    self.style.WARNING(
                        "Could not find a suitable release of '{package_name}'".format(
                            package_name=package_name
                        )
                    )
                )

            return latest

        self.stdout.write(
            self.style.WARNING(
                "Package '{package_name}' is not registered in PyPI".format(
                    package_name=package_name
                )
            )
        )

        return None

    def _import_from_imicusfat(self) -> None:
        """
        Start the import
        :return:
        :rtype:
        """

        # First, we check if the target tables are empty ...
        current_afat_count = AFat.objects.all().count()
        current_afat_links_count = AFatLink.objects.all().count()
        current_afat_linktype_count = AFatLinkType.objects.all().count()
        current_afat_clickduration_count = ClickAFatDuration.objects.all().count()

        if (
            current_afat_count > 0
            or current_afat_links_count > 0
            or current_afat_linktype_count > 0
            or current_afat_clickduration_count > 0
        ):
            self.stdout.write(
                self.style.WARNING(
                    "You already have FAT data with the aFAT module. "
                    "Import cannot be continued."
                )
            )

            return

        # Import fat link type
        imicusfat_fleettypes = IFatLinkType.objects.all()
        for imicusfat_fleettype in imicusfat_fleettypes:
            self.stdout.write(
                "Importing fleet type '{fleet_type}'.".format(
                    fleet_type=imicusfat_fleettype.name
                )
            )

            afat_fleettype = AFatLinkType()

            afat_fleettype.id = imicusfat_fleettype.id
            afat_fleettype.name = imicusfat_fleettype.name
            afat_fleettype.is_enabled = imicusfat_fleettype.is_enabled

            afat_fleettype.save()

        # Import FAT links
        imicusfat_fatlinks = IFatLink.objects.all()
        for imicusfat_fatlink in imicusfat_fatlinks:
            self.stdout.write(
                "Importing FAT link for fleet '{fleet}' with "
                "hash '{fatlink_hash}'.".format(
                    fleet=imicusfat_fatlink.fleet,
                    fatlink_hash=imicusfat_fatlink.hash,
                )
            )

            afatlink = AFatLink()

            afatlink.id = imicusfat_fatlink.id
            afatlink.afattime = imicusfat_fatlink.ifattime
            afatlink.fleet = (
                imicusfat_fatlink.fleet
                if imicusfat_fatlink.fleet is not None
                else imicusfat_fatlink.hash
            )
            afatlink.hash = imicusfat_fatlink.hash
            afatlink.creator_id = imicusfat_fatlink.creator_id
            afatlink.link_type_id = imicusfat_fatlink.link_type_id
            afatlink.is_esilink = imicusfat_fatlink.is_esilink

            afatlink.save()

            # Write to log table
            if imicusfat_fatlink.is_esilink:
                log_text = (
                    "ESI FAT link {fatlink_hash} with name {name} was created by {user}"
                ).format(
                    fatlink_hash=imicusfat_fatlink.hash,
                    name=imicusfat_fatlink.fleet,
                    user=imicusfat_fatlink.creator,
                )
            else:
                try:
                    fleet_duration = ClickIFatDuration.objects.get(
                        fleet_id=imicusfat_fatlink.id
                    )

                    log_text = (
                        "FAT link {fatlink_hash} with name {name} and a "
                        "duration of {duration} minutes was created by {user}"
                    ).format(
                        fatlink_hash=imicusfat_fatlink.hash,
                        name=imicusfat_fatlink.fleet,
                        duration=fleet_duration.duration,
                        user=imicusfat_fatlink.creator,
                    )
                except ClickIFatDuration.DoesNotExist:
                    log_text = (
                        "FAT link {fatlink_hash} with name {name} was created by {user}"
                    ).format(
                        fatlink_hash=imicusfat_fatlink.hash,
                        name=imicusfat_fatlink.fleet,
                        user=imicusfat_fatlink.creator,
                    )

            afatlog = AFatLog()
            afatlog.log_time = imicusfat_fatlink.ifattime
            afatlog.log_event = AFatLogEvent.CREATE_FATLINK
            afatlog.log_text = log_text
            afatlog.user_id = imicusfat_fatlink.creator_id
            afatlog.save()

        # Import FATs
        imicustaf_fats = IFat.objects.all()
        for imicusfat_fat in imicustaf_fats:
            self.stdout.write(
                "Importing FATs for FAT link ID '{fatlink_id}'.".format(
                    fatlink_id=imicusfat_fat.id
                )
            )

            afat = AFat()

            afat.id = imicusfat_fat.id
            afat.system = imicusfat_fat.system
            afat.shiptype = imicusfat_fat.shiptype
            afat.character_id = imicusfat_fat.character_id
            afat.afatlink_id = imicusfat_fat.ifatlink_id

            afat.save()

        # Import click FAT durations
        imicusfat_clickfatdurations = ClickIFatDuration.objects.all()
        for imicusfat_clickfatduration in imicusfat_clickfatdurations:
            self.stdout.write(
                "Importing FAT duration with ID '{duration_id}'.".format(
                    duration_id=imicusfat_clickfatduration.id
                )
            )

            afat_clickfatduration = ClickAFatDuration()

            afat_clickfatduration.id = imicusfat_clickfatduration.id
            afat_clickfatduration.duration = imicusfat_clickfatduration.duration
            afat_clickfatduration.fleet_id = imicusfat_clickfatduration.fleet_id

            afat_clickfatduration.save()

        # Import manual fat to log table
        imicusfat_manualfats = ManualIFat.objects.all()
        for imicusfat_manualfat in imicusfat_manualfats:
            self.stdout.write(
                "Importing manual FAT with ID '{manualfat_id}'.".format(
                    manualfat_id=imicusfat_manualfat.id
                )
            )

            fatlink = IFatLink.objects.get(manualifat=imicusfat_manualfat)
            log_text = (
                "Pilot {pilot_name} was manually added to "
                'FAT link with hash "{fatlink_hash}"'
            ).format(
                pilot_name=imicusfat_manualfat.character.character_name,
                fatlink_hash=fatlink.hash,
            )

            if imicusfat_manualfat.created_at is not None:
                afatlog = AFatLog()
                afatlog.log_time = imicusfat_manualfat.created_at
                afatlog.log_event = AFatLogEvent.MANUAL_FAT
                afatlog.log_text = log_text
                afatlog.user_id = imicusfat_manualfat.creator_id
                afatlog.save()

        self.stdout.write(
            self.style.SUCCESS(
                "Import complete! "
                "You can now deactivate the ImicusFAT module in your local.py"
            )
        )

    def handle(self, *args, **options):
        """
        Ask before running ...
        :param args:
        :type args:
        :param options:
        :type options:
        :return:
        :rtype:
        """

        if imicusfat_installed():
            has_conflict = False

            self.stdout.write(
                self.style.SUCCESS("ImicusFAT module is active, let's go!")
            )

            self.stdout.write("Checking for potentially available updates ...")

            ifat_version_available = self.latest_version_available(
                package_name="allianceauth-imicusfat"
            )

            afat_version_available = self.latest_version_available(
                package_name="allianceauth-afat"
            )

            # Check if updates for ImicusFAT are available
            if ifat_version_available is not None:
                if version_parse(ifat_version_installed) < version_parse(
                    ifat_version_available
                ):
                    self.stdout.write(
                        self.style.WARNING(
                            "ImicusFAT is outdated. "
                            "Please update to the latest ImicusFAT version first."
                        )
                    )

                    self.stdout.write(
                        self.style.WARNING(
                            "ImicusFAT version installed: "
                            "{ifat_version_installed}".format(
                                ifat_version_installed=ifat_version_installed
                            )
                        )
                    )

                    self.stdout.write(
                        self.style.WARNING(
                            "ImicusFAT version available: "
                            "{ifat_version_available}".format(
                                ifat_version_available=ifat_version_available
                            )
                        )
                    )

                    has_conflict = True
            else:
                has_conflict = True

            # Check if updates for aFAT are available
            if afat_version_available is not None:
                if version_parse(afat_version_installed) < version_parse(
                    afat_version_available
                ):
                    self.stdout.write(
                        self.style.WARNING(
                            "aFAT is outdated. "
                            "Please update to the latest aFAT version first."
                        )
                    )

                    self.stdout.write(
                        self.style.WARNING(
                            "aFAT version installed: {afat_version_installed}".format(
                                afat_version_installed=afat_version_installed
                            )
                        )
                    )

                    self.stdout.write(
                        self.style.WARNING(
                            "aFAT version available: {afat_version_available}".format(
                                afat_version_available=afat_version_available
                            )
                        )
                    )

                    has_conflict = True
            else:
                has_conflict = True

            if has_conflict is False:
                self.stdout.write(
                    "Importing all FAT/FAT link data from ImicusFAT module. "
                    "This can only be done once during the very first installation. "
                    "As soon as you have data collected with your AFAT module, "
                    "this import will fail!"
                )

                user_input = get_input("Are you sure you want to proceed? (yes/no)?")

                if user_input == "yes":
                    self.stdout.write("Starting import. Please stand by.")
                    self._import_from_imicusfat()
                else:
                    self.stdout.write(self.style.WARNING("Aborted."))
        else:
            self.stdout.write(
                self.style.WARNING(
                    "ImicusFAT module is not active. "
                    "Please make sure you have it in your "
                    "INSTALLED_APPS in your local.py! "
                    "Aborting."
                )
            )



"""
Import FAT data from bFAT module
"""

from bfat.models import ClickFatDuration as BfatClickFatDuration
from bfat.models import Fat as BfatFat
from bfat.models import FatLink as BfatFatLink
from bfat.models import ManualFat as BfatManualFat

from django.apps import apps
from django.core.management.base import BaseCommand

from afat.models import AFat, AFatLink, AFatLog, AFatLogEvent, ClickAFatDuration


def get_input(text) -> str:
    """
    Wrapped input to enable import
    :param text:
    :type text:
    :return:
    :rtype:
    """

    return input(text)


def bfat_installed() -> bool:
    """
    Check if bfat is installed
    :return:
    :rtype:
    """

    return apps.is_installed("bfat")


class Command(BaseCommand):
    """
    Initial import of FAT data from AA FAT module
    """

    help = "Imports FAT data from ImicusFAT module"

    def _import_from_imicusfat(self) -> None:
        """
        Start the import
        :return:
        :rtype:
        """

        # Check if AA FAT is active
        if bfat_installed():
            self.stdout.write(
                self.style.SUCCESS("ImicusFAT module is active, let's go!")
            )

            # First, we check if the target tables are empty ...
            current_afat_count = AFat.objects.all().count()
            current_afat_links_count = AFatLink.objects.all().count()
            current_afat_clickduration_count = ClickAFatDuration.objects.all().count()

            if (
                current_afat_count > 0
                or current_afat_links_count > 0
                or current_afat_clickduration_count > 0
            ):
                self.stdout.write(
                    self.style.WARNING(
                        "You already have FAT data with the AFAT module. "
                        "Import cannot be continued."
                    )
                )

                return

            # Import FAT links
            bfat_fatlinks = BfatFatLink.objects.all()
            for bfat_fatlink in bfat_fatlinks:
                self.stdout.write(
                    f'Importing FAT link for fleet "{bfat_fatlink.fleet}" '
                    f'with hash "{bfat_fatlink.hash}".'
                )

                afatlink = AFatLink()

                afatlink.id = bfat_fatlink.id
                afatlink.afattime = bfat_fatlink.fattime
                afatlink.fleet = (
                    bfat_fatlink.fleet
                    if bfat_fatlink.fleet is not None
                    else bfat_fatlink.hash
                )
                afatlink.hash = bfat_fatlink.hash
                afatlink.creator_id = bfat_fatlink.creator_id

                afatlink.save()

                # Write to log table
                try:
                    fleet_duration = BfatClickFatDuration.objects.get(
                        fleet_id=bfat_fatlink.id
                    )

                    log_text = (
                        f'FAT link "{bfat_fatlink.hash}" with name '
                        f'"{bfat_fatlink.fleet}" and a duration of '
                        f"{fleet_duration.duration} minutes was created "
                        f"by {bfat_fatlink.creator}"
                    )
                except BfatClickFatDuration.DoesNotExist:
                    log_text = (
                        f'FAT link "{bfat_fatlink.hash}" with name '
                        f'"{bfat_fatlink.fleet}" was created by {bfat_fatlink.creator}'
                    )

                if bfat_fatlink.fattime is not None:
                    afatlog = AFatLog()
                    afatlog.log_time = bfat_fatlink.fattime
                    afatlog.log_event = AFatLogEvent.CREATE_FATLINK
                    afatlog.log_text = log_text
                    afatlog.user_id = bfat_fatlink.creator_id
                    afatlog.save()

            # Import FATs
            bfat_fats = BfatFat.objects.all()
            for bfat_fat in bfat_fats:
                self.stdout.write(f"Importing FATs for FAT link ID {bfat_fat.id}.")

                afat = AFat()

                afat.id = bfat_fat.id
                afat.system = bfat_fat.system
                afat.shiptype = bfat_fat.shiptype
                afat.character_id = bfat_fat.character_id
                afat.afatlink_id = bfat_fat.fatlink_id

                afat.save()

            # Import click FAT durations
            bfat_clickfatdurations = BfatClickFatDuration.objects.all()
            for bfat_clickfatduration in bfat_clickfatdurations:
                self.stdout.write(
                    f"Importing FAT duration with ID {bfat_clickfatduration.id}."
                )

                afat_clickfatduration = ClickAFatDuration()

                afat_clickfatduration.id = bfat_clickfatduration.id
                afat_clickfatduration.duration = bfat_clickfatduration.duration
                afat_clickfatduration.fleet_id = bfat_clickfatduration.fleet_id

                afat_clickfatduration.save()

            # Import manual fat
            bfat_manualfats = BfatManualFat.objects.all()
            for bfat_manualfat in bfat_manualfats:
                self.stdout.write(f"Importing manual FAT with ID {bfat_manualfat.id}.")

                fatlink = BfatFatLink.objects.get(manualfat=bfat_manualfat)
                log_text = (
                    f"Pilot {bfat_manualfat.character.character_name} was manually "
                    f'added to FAT link with hash "{fatlink.hash}"'
                )

                afatlog = AFatLog()
                afatlog.log_time = bfat_manualfat.created_at
                afatlog.log_event = AFatLogEvent.MANUAL_FAT
                afatlog.log_text = log_text
                afatlog.user_id = bfat_manualfat.creator_id
                afatlog.save()

            self.stdout.write(
                self.style.SUCCESS(
                    "Import complete! "
                    "You can now deactivate the bFAT module in your local.py"
                )
            )
        else:
            self.stdout.write(
                self.style.WARNING(
                    "bFAT module is not active. "
                    "Please make sure you have it in your "
                    "INSTALLED_APPS in your local.py!"
                )
            )

    def handle(self, *args, **options):
        """
        Ask before running ...
        :param args:
        :type args:
        :param options:
        :type options:
        :return:
        :rtype:
        """

        self.stdout.write(
            "Importing all FAT/FAT link data from bFAT module. "
            "This can only be done once during the very first installation. "
            "As soon as you have data collected with your AFAT module, "
            "this import will fail!"
        )

        user_input = get_input("Are you sure you want to proceed? (yes/no)?")

        if user_input == "yes":
            self.stdout.write("Starting import. Please stand by.")
            self._import_from_imicusfat()
        else:
            self.stdout.write(self.style.WARNING("Aborted."))



# Generated by Django 3.2.9 on 2021-11-09 22:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("afat", "0020_auto_20210617_1037"),
    ]

    operations = [
        migrations.AlterField(
            model_name="afatlink",
            name="fleet",
            field=models.CharField(
                default=None, help_text="The FAT link fleet name", max_length=254
            ),
        ),
    ]



# Generated by Django 2.0.6 on 2018-06-06 07:57

import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import afat.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0010_alliance_ticker"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="AFat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("system", models.CharField(max_length=30)),
                ("shiptype", models.CharField(max_length=30)),
                ("station", models.CharField(max_length=125)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.EveCharacter",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AFatLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("afattime", models.DateTimeField(default=django.utils.timezone.now)),
                ("fleet", models.CharField(max_length=254)),
                ("hash", models.CharField(max_length=254)),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=models.SET(afat.models.get_sentinel_user),
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="afat",
            name="afatlink",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="afat.AFatLink"
            ),
        ),
        migrations.AddField(
            model_name="afat",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterUniqueTogether(
            name="afat",
            unique_together={("character", "afatlink")},
        ),
    ]



"""
Test fatlinks views
"""

import datetime as dt

# import json
from datetime import datetime

from pytz import utc

from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ..models import AFat, AFatLink, AFatLinkType, ClickAFatDuration, get_hash_on_save
from ..utils import get_main_character_from_user
from .fixtures.load_allianceauth import load_allianceauth

MODULE_PATH = "afat.views.fatlinks"


class TestFatlinksView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_without_access, _ = create_user_from_evecharacter(
            cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            cls.character_1002.character_id, permissions=["afat.basic_access"]
        )

        cls.user_with_manage_afat, _ = create_user_from_evecharacter(
            cls.character_1003.character_id,
            permissions=["afat.basic_access", "afat.manage_afat"],
        )

        cls.user_with_add_fatlink, _ = create_user_from_evecharacter(
            cls.character_1004.character_id,
            permissions=["afat.basic_access", "afat.add_fatlink"],
        )

        # cls.afat_link_type_cta = AFatLinkType.objects.create(name="CTA")
        # cls.afat_link_type_stratop = AFatLinkType.objects.create(name="Strat OP")

        # Generate some FAT links and FATs
        cls.afat_link_april_1 = AFatLink.objects.create(
            fleet="April Fleet 1",
            hash="1231",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            afattime=dt.datetime(2020, 4, 1, tzinfo=utc),
        )
        cls.afat_link_april_2 = AFatLink.objects.create(
            fleet="April Fleet 2",
            hash="1232",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            afattime=dt.datetime(2020, 4, 15, tzinfo=utc),
        )
        cls.afat_link_september = AFatLink.objects.create(
            fleet="September Fleet",
            hash="1233",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            afattime=dt.datetime(2020, 9, 1, tzinfo=utc),
        )
        cls.afat_link_september_no_fats = AFatLink.objects.create(
            fleet="September Fleet 2",
            hash="1234",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            afattime=dt.datetime(2020, 9, 1, tzinfo=utc),
        )

        AFat.objects.create(
            character=cls.character_1101,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1001,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1002,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1003,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1004,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1005,
            afatlink=cls.afat_link_april_1,
            shiptype="Omen",
        )

        AFat.objects.create(
            character=cls.character_1001,
            afatlink=cls.afat_link_april_2,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1004,
            afatlink=cls.afat_link_april_2,
            shiptype="Thorax",
        )
        AFat.objects.create(
            character=cls.character_1002,
            afatlink=cls.afat_link_april_2,
            shiptype="Thorax",
        )
        AFat.objects.create(
            character=cls.character_1003,
            afatlink=cls.afat_link_april_2,
            shiptype="Omen",
        )

        AFat.objects.create(
            character=cls.character_1001,
            afatlink=cls.afat_link_september,
            shiptype="Omen",
        )
        AFat.objects.create(
            character=cls.character_1004,
            afatlink=cls.afat_link_september,
            shiptype="Guardian",
        )
        AFat.objects.create(
            character=cls.character_1002,
            afatlink=cls.afat_link_september,
            shiptype="Omen",
        )

    def test_should_show_fatlnks_overview(self):
        # given
        self.client.force_login(self.user_with_basic_access)

        # when
        url = reverse("afat:fatlinks_overview")
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_fatlnks_overview_with_year(self):
        # given
        self.client.force_login(self.user_with_basic_access)

        # when
        url = reverse("afat:fatlinks_overview", kwargs={"year": 2020})
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_add_fatlink_for_user_with_manage_afat(self):
        # given
        AFatLinkType.objects.create(name="CTA")

        self.client.force_login(self.user_with_manage_afat)

        # when
        url = reverse("afat:fatlinks_add_fatlink")
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_add_fatlink_for_user_with_add_fatlinkt(self):
        # given
        self.client.force_login(self.user_with_add_fatlink)

        # when
        url = reverse("afat:fatlinks_add_fatlink")
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_fatlink_details_for_user_with_manage_afat(self):
        # given
        self.client.force_login(self.user_with_manage_afat)

        # when
        url = reverse(
            "afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": self.afat_link_april_1.hash},
        )
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_show_fatlink_details_for_user_with_add_fatlinkt(self):
        # given
        self.client.force_login(self.user_with_add_fatlink)

        # when
        url = reverse(
            "afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": self.afat_link_april_1.hash},
        )
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_not_show_fatlink_details_for_non_existing_fatlink(self):
        # given
        self.client.force_login(self.user_with_manage_afat)

        # when
        url = reverse(
            "afat:fatlinks_details_fatlink",
            kwargs={"fatlink_hash": "foobarsson"},
        )
        res = self.client.get(url)

        # then
        self.assertNotEqual(res.status_code, 200)
        self.assertEqual(res.status_code, 302)

        messages = list(get_messages(res.wsgi_request))

        self.assertRaises(AFatLink.DoesNotExist)
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            str(messages[0]),
            "<h4>Warning!</h4><p>The hash provided is not valid.</p>",
        )

    def test_ajax_get_fatlinks_by_year(self):
        # given
        self.maxDiff = None
        self.client.force_login(self.user_with_basic_access)

        fatlink_hash = get_hash_on_save()
        fatlink_type_cta = AFatLinkType.objects.create(name="CTA")
        fatlink_created = AFatLink.objects.create(
            fleet="April Fleet 1",
            creator=self.user_with_manage_afat,
            character=self.character_1001,
            hash=fatlink_hash,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id=3726458287,
            link_type=fatlink_type_cta,
            afattime="2021-11-05T13:19:49.676Z",
        )

        ClickAFatDuration.objects.create(fleet=fatlink_created, duration=120)

        # when
        fatlink = (
            AFatLink.objects.select_related_default()
            .annotate_afats_count()
            .get(hash=fatlink_hash)
        )

        url_with_year = reverse(
            "afat:fatlinks_ajax_get_fatlinks_by_year",
            kwargs={"year": datetime.now().year},
        )
        result = self.client.get(url_with_year)

        # then
        self.assertEqual(result.status_code, 200)

        creator_main_character = get_main_character_from_user(user=fatlink.creator)
        fleet_time = fatlink.afattime
        fleet_time_timestamp = fleet_time.timestamp()
        esi_marker = (
            '<span class="label label-default afat-label afat-label-via-esi '
            'afat-label-active-esi-fleet">via ESI</span>'
        )

        self.assertJSONEqual(
            str(result.content, encoding="utf8"),
            [
                {
                    "pk": fatlink.pk,
                    "fleet_name": fatlink.fleet + esi_marker,
                    "creator_name": creator_main_character,
                    "fleet_type": "CTA",
                    "fleet_time": {
                        "time": "2021-11-05T13:19:49.676Z",
                        "timestamp": fleet_time_timestamp,
                    },
                    "fats_number": 0,
                    "hash": fatlink.hash,
                    "is_esilink": True,
                    "esi_fleet_id": fatlink.esi_fleet_id,
                    "is_registered_on_esi": True,
                    # "actions": '<a class="btn btn-afat-action btn-info btn-sm" href="/fleet-activity-tracking/fatlink/ncOsHjnjmYZd9k6hI4us8QShRlqJ17/details/"><span class="fas fa-eye"></span></a><a class="btn btn-afat-action btn-danger btn-sm" data-toggle="modal" data-target="#deleteFatLinkModal" data-url="/fleet-activity-tracking/fatlink/ncOsHjnjmYZd9k6hI4us8QShRlqJ17/delete/" data-confirm-text="Delete"data-body-text="<p>Are you sure you want to delete FAT link M2-XFE Keepstar Kill?</p>"><span class="glyphicon glyphicon-trash"></span></a>',
                    "actions": "",
                    "via_esi": "Yes",
                }
            ],
        )



# Generated by Django 3.1.8 on 2021-05-01 22:25

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0014_auto_20210105_1413"),
        ("afat", "0018_auto_20210420_2016"),
    ]

    operations = [
        migrations.AlterField(
            model_name="afat",
            name="afatlink",
            field=models.ForeignKey(
                help_text="The fatlink the character registered at",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afats",
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="character",
            field=models.ForeignKey(
                help_text="Character who registered this fat",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afats",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="shiptype",
            field=models.CharField(
                db_index=True,
                help_text="The ship the character was flying",
                max_length=100,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="afattime",
            field=models.DateTimeField(
                db_index=True,
                default=django.utils.timezone.now,
                help_text="When was this fatlink created",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="hash",
            field=models.CharField(
                db_index=True, help_text="The fatlinks hash", max_length=254
            ),
        ),
        migrations.AlterField(
            model_name="afatlog",
            name="log_time",
            field=models.DateTimeField(
                db_index=True, default=django.utils.timezone.now
            ),
        ),
    ]



"""
Our app setting
"""

from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _

from app_utils.django import clean_setting

# Set default expiry time in minutes
AFAT_DEFAULT_FATLINK_EXPIRY_TIME = clean_setting("AFAT_DEFAULT_FATLINK_EXPIRY_TIME", 60)

# Set the default time in minutes a FAT lnk can be re-opened after it is expired
AFAT_DEFAULT_FATLINK_REOPEN_GRACE_TIME = clean_setting(
    "AFAT_DEFAULT_FATLINK_REOPEN_GRACE_TIME", 60
)

# Set the default time in minutes a FAT link is re-opened for
AFAT_DEFAULT_FATLINK_REOPEN_DURATION = clean_setting(
    "AFAT_DEFAULT_FATLINK_REOPEN_DURATION", 60
)

AFAT_DEFAULT_LOG_DURATION = clean_setting("AFAT_DEFAULT_LOG_DURATION", 60)

# Name of this app, as shown in the Auth sidebar and page titles
AFAT_APP_NAME = clean_setting(
    "AFAT_APP_NAME", _("Fleet Activity Tracking"), required_type=str
)

AFAT_BASE_URL = slugify(AFAT_APP_NAME, allow_unicode=True)



"""
views helper
"""

import random

from django.contrib.auth.models import Permission
from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext as _

from allianceauth.eveonline.models import EveCharacter
from app_utils.django import users_with_permission

from afat.models import AFat, AFatLink, AFatLog, AFatLogEvent
from afat.utils import get_main_character_from_user


def convert_fatlinks_to_dict(
    request: WSGIRequest, fatlink: AFatLink, close_esi_redirect: str = None
) -> dict:
    """
    Converts an AFatLink object into a dictionary
    :param request:
    :type request:
    :param fatlink:
    :type fatlink:
    :param close_esi_redirect:
    :type close_esi_redirect:
    :return:
    :rtype:
    """

    # Fleet name
    fatlink_fleet = fatlink.fleet if fatlink.fleet is not None else fatlink.hash

    # ESI marker
    via_esi = "No"
    esi_fleet_marker = ""

    # Check for ESI link
    if fatlink.is_esilink:
        via_esi = "Yes"
        esi_fleet_marker_classes = "label label-default afat-label afat-label-via-esi"

        if fatlink.is_registered_on_esi:
            esi_fleet_marker_classes += " afat-label-active-esi-fleet"

        marker_text = _("via ESI")
        esi_fleet_marker += (
            f'<span class="{esi_fleet_marker_classes}">{marker_text}</span>'
        )

    # Fleet type
    fatlink_type = fatlink.link_type.name if fatlink.link_type else ""

    # Creator name
    creator_main_character = get_main_character_from_user(user=fatlink.creator)

    # Fleet time
    fleet_time = fatlink.afattime
    fleet_time_timestamp = fleet_time.timestamp()

    # Action buttons
    actions = ""
    if (
        fatlink.is_esilink
        and fatlink.is_registered_on_esi
        and fatlink.creator == request.user
    ):
        button_close_esi_tracking_url = reverse(
            "afat:fatlinks_close_esi_fatlink", args=[fatlink.hash]
        )

        close_esi_redirect_parameter = (
            f"?next={close_esi_redirect}" if close_esi_redirect is not None else ""
        )

        button_title = _(
            "Clicking here will stop the automatic tracking through ESI for this fleet and close the associated FAT link."
        )
        modal_body_text = _(
            f"<p>Are you sure you want to close ESI fleet with ID {fatlink.esi_fleet_id} from {fatlink.character.character_name}?</p>"
        )
        modal_confirm_text = _("Stop Tracking")

        actions += (
            '<a class="btn btn-afat-action btn-primary btn-sm" '
            f'style="margin-left: 0.25rem;" title="{button_title}" data-toggle="modal" '
            'data-target="#cancelEsiFleetModal" '
            f'data-url="{button_close_esi_tracking_url}{close_esi_redirect_parameter}" '
            f'data-body-text="{modal_body_text}" '
            f'data-confirm-text="{modal_confirm_text}">'
            '<i class="fas fa-times"></i></a>'
        )

    if request.user.has_perm("afat.manage_afat") or request.user.has_perm(
        "afat.add_fatlink"
    ):
        button_edit_url = reverse("afat:fatlinks_details_fatlink", args=[fatlink.hash])

        actions += (
            '<a class="btn btn-afat-action btn-info btn-sm" '
            f'href="{button_edit_url}"><span class="fas fa-eye"></span></a>'
        )

    if request.user.has_perm("afat.manage_afat"):
        button_delete_url = reverse("afat:fatlinks_delete_fatlink", args=[fatlink.hash])
        button_delete_text = _("Delete")
        modal_body_text = _(
            f"<p>Are you sure you want to delete FAT link {fatlink_fleet}?</p>"
        )

        actions += (
            '<a class="btn btn-afat-action btn-danger btn-sm" data-toggle="modal" '
            f'data-target="#deleteFatLinkModal" data-url="{button_delete_url}" '
            f'data-confirm-text="{button_delete_text}"'
            f'data-body-text="{modal_body_text}">'
            '<span class="glyphicon glyphicon-trash">'
            "</span></a>"
        )

    summary = {
        "pk": fatlink.pk,
        "fleet_name": fatlink_fleet + esi_fleet_marker,
        "creator_name": creator_main_character,
        "fleet_type": fatlink_type,
        "fleet_time": {"time": fleet_time, "timestamp": fleet_time_timestamp},
        "fats_number": fatlink.afats_count,
        "hash": fatlink.hash,
        "is_esilink": fatlink.is_esilink,
        "esi_fleet_id": fatlink.esi_fleet_id,
        "is_registered_on_esi": fatlink.is_registered_on_esi,
        "actions": actions,
        "via_esi": via_esi,
    }

    return summary


def convert_fats_to_dict(request: WSGIRequest, fat: AFat) -> dict:
    """
    Converts an AFat object into a dictionary
    :param request:
    :type request:
    :param fat:
    :type fat:
    :return:
    :rtype:
    """

    # fleet type
    fleet_type = (
        fat.afatlink.link_type.name if fat.afatlink.link_type is not None else ""
    )

    # esi marker
    via_esi = "No"
    esi_fleet_marker = ""

    if fat.afatlink.is_esilink:
        via_esi = "Yes"
        esi_fleet_marker_classes = "label label-default afat-label afat-label-via-esi"

        if fat.afatlink.is_registered_on_esi:
            esi_fleet_marker_classes += " afat-label-active-esi-fleet"

        marker_text = _("via ESI")
        esi_fleet_marker += (
            f'<span class="{esi_fleet_marker_classes}">{marker_text}</span>'
        )

    # actions
    actions = ""
    if request.user.has_perm("afat.manage_afat"):
        button_delete_fat = reverse(
            "afat:fatlinks_delete_fat", args=[fat.afatlink.hash, fat.id]
        )
        button_delete_text = _("Delete")
        modal_body_text = _(
            f"<p>Are you sure you want to remove {fat.character.character_name} from this FAT link?</p>"
        )

        actions += (
            '<a class="btn btn-danger btn-sm" '
            'data-toggle="modal" '
            'data-target="#deleteFatModal" '
            f'data-url="{button_delete_fat}" '
            f'data-confirm-text="{button_delete_text}"'
            f'data-body-text="{modal_body_text}">'
            '<span class="glyphicon glyphicon-trash"></span>'
            "</a>"
        )

    fleet_time = fat.afatlink.afattime
    fleet_time_timestamp = fleet_time.timestamp()
    fleet_name = (
        fat.afatlink.fleet if fat.afatlink.fleet is not None else fat.afatlink.hash
    )

    summary = {
        "system": fat.system,
        "ship_type": fat.shiptype,
        "character_name": fat.character.character_name,
        "fleet_name": fleet_name + esi_fleet_marker,
        "fleet_time": {"time": fleet_time, "timestamp": fleet_time_timestamp},
        "fleet_type": fleet_type,
        "via_esi": via_esi,
        "actions": actions,
    }

    return summary


def convert_logs_to_dict(log: AFatLog, fatlink_exists: bool = False) -> dict:
    """
    Convert AFatLog to dict
    :param log:
    :type log:
    :param fatlink_exists:
    :type fatlink_exists:
    :return:
    :rtype:
    """

    log_time = log.log_time
    log_time_timestamp = log_time.timestamp()

    # User name
    user_main_character = get_main_character_from_user(user=log.user)

    fatlink_html = _(f"{log.fatlink_hash} (Deleted)")
    if fatlink_exists is True:
        fatlink_link = reverse("afat:fatlinks_details_fatlink", args=[log.fatlink_hash])
        fatlink_html = f'<a href="{fatlink_link}">{log.fatlink_hash}</a>'

    fatlink = {"html": fatlink_html, "hash": log.fatlink_hash}

    summary = {
        "log_time": {"time": log_time, "timestamp": log_time_timestamp},
        "log_event": AFatLogEvent(log.log_event).label,
        "user": user_main_character,
        "fatlink": fatlink,
        "description": log.log_text,
    }

    return summary


def get_random_rgba_color():
    """
    Get a random RGB(a) color
    :return:
    :rtype:
    """

    return "rgba({red}, {green}, {blue}, 1)".format(
        red=random.randint(0, 255),
        green=random.randint(0, 255),
        blue=random.randint(0, 255),
    )


def characters_with_permission(permission: Permission) -> models.QuerySet:
    """
    Returns queryset of characters that have the given permission
    in Auth through due to their associated user
    :param permission:
    :type permission:
    :return:
    :rtype:
    """

    # First we need the users that have the permission
    users_qs = users_with_permission(permission)

    # Now get their characters ...
    charater_qs = EveCharacter.objects.filter(character_ownership__user__in=users_qs)

    return charater_qs



"""
Test access to the afat module
"""

from django.test import TestCase
from django.urls import reverse

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from .fixtures.load_allianceauth import load_allianceauth

MODULE_PATH = "afat.views.statistics"


class TestAccesss(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)

        cls.user_without_access, _ = create_user_from_evecharacter(
            cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            cls.character_1002.character_id, permissions=["afat.basic_access"]
        )

    def test_should_show_afat_dashboard_for_user_with_basic_access(self):
        # given
        self.client.force_login(self.user_with_basic_access)

        # when
        url = reverse("afat:dashboard")
        res = self.client.get(url)

        # then
        self.assertEqual(res.status_code, 200)

    def test_should_not_show_afat_dashboard_for_user_without_access(self):
        # given
        self.client.force_login(self.user_without_access)

        # when
        url = reverse("afat:dashboard")
        res = self.client.get(url)

        # then
        self.assertNotEqual(res.status_code, 200)
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, "/account/login/?next=/fleet-activity-tracking/")



"""
Tasks
"""

from datetime import timedelta

from bravado.exception import (
    HTTPBadGateway,
    HTTPGatewayTimeout,
    HTTPNotFound,
    HTTPServiceUnavailable,
)
from celery import shared_task

from django.core.cache import cache
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.logging import LoggerAddTag
from esi.models import Token

from afat import __title__
from afat.app_settings import AFAT_DEFAULT_LOG_DURATION
from afat.models import AFat, AFatLink, AFatLog
from afat.providers import esi
from afat.utils import get_or_create_character

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


DEFAULT_TASK_PRIORITY = 6

ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_MAX_RETRIES = 3

CACHE_KEY_FLEET_CHANGED_ERROR = (
    "afat_task_update_esi_fatlinks_error_counter_fleet_changed_"
)
CACHE_KEY_NOT_IN_FLEET_ERROR = (
    "afat_task_update_esi_fatlinks_error_counter_not_in_fleet_"
)
CACHE_KEY_NO_FLEET_ERROR = "afat_task_update_esi_fatlinks_error_counter_no_fleet_"
CACHE_KEY_NO_FLEETBOSS_ERROR = (
    "afat_task_update_esi_fatlinks_error_counter_no_fleetboss_"
)
CACHE_MAX_ERROR_COUNT = 3

# Params for all tasks
TASK_DEFAULT_KWARGS = {
    "time_limit": 120,  # Stop after 2 minutes
}

# Params for tasks that make ESI calls
TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": ESI_MAX_RETRIES},
        "retry_backoff": True,
    },
}


@shared_task
def process_fats(data_list, data_source, fatlink_hash):
    """
    Due to the large possible size of fatlists,
    this process will be scheduled to process esi data
    and possible other sources in the future.
    :param data_list:
    :type data_list:
    :param data_source:
    :type data_source:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    if data_source == "esi":
        logger.info(
            f'Valid fleet for FAT link hash "{fatlink_hash}" found '
            f"registered via ESI, checking for new pilots"
        )

        for char in data_list:
            process_character.delay(char, fatlink_hash)


@shared_task
def process_character(char, fatlink_hash):
    """
    Process character
    :param char:
    :type char:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    link = AFatLink.objects.get(hash=fatlink_hash)
    char_id = char["character_id"]
    character = get_or_create_character(character_id=char_id)

    # Only process if the character is not already registered for this FAT
    if AFat.objects.filter(character=character, afatlink_id=link.pk).exists() is False:
        solar_system_id = char["solar_system_id"]
        ship_type_id = char["ship_type_id"]

        solar_system = esi.client.Universe.get_universe_systems_system_id(
            system_id=solar_system_id
        ).result()
        ship = esi.client.Universe.get_universe_types_type_id(
            type_id=ship_type_id
        ).result()

        solar_system_name = solar_system["name"]
        ship_name = ship["name"]

        logger.info(
            "New Pilot: Adding {character_name} in {system_name} flying a {ship_name} "
            'to FAT link "{fatlink_hash}"'.format(
                character_name=character,
                system_name=solar_system_name,
                ship_name=ship_name,
                fatlink_hash=fatlink_hash,
            )
        )

        AFat(
            afatlink_id=link.pk,
            character=character,
            system=solar_system_name,
            shiptype=ship_name,
        ).save()


def close_esi_fleet(fatlink: AFatLink, reason: str) -> None:
    """
    Closing ESI fleet
    :param fatlink:
    :type fatlink:
    :param reason:
    :type reason:
    :return:
    :rtype:
    """

    logger.info(
        'Closing ESI FAT link with hash "{fatlink_hash}". Reason: {reason}'.format(
            fatlink_hash=fatlink.hash, reason=reason
        )
    )

    fatlink.is_registered_on_esi = False
    fatlink.save()


def esi_fatlinks_error_handling(
    cache_key: str, fatlink: AFatLink, logger_message: str
) -> None:
    """
    ESI error handling
    :param cache_key:
    :type cache_key:
    :param fatlink:
    :type fatlink:
    :param logger_message:
    :type logger_message:
    :return:
    :rtype:
    """

    if int(cache.get(cache_key + fatlink.hash)) < CACHE_MAX_ERROR_COUNT:
        error_count = int(cache.get(cache_key + fatlink.hash))

        error_count += 1

        logger.info(
            (
                'FAT link "{falink_hash}" Error: "{logger_message}" '
                "({error_count} of {max_count})."
            ).format(
                falink_hash=fatlink.hash,
                logger_message=logger_message,
                error_count=error_count,
                max_count=CACHE_MAX_ERROR_COUNT,
            )
        )

        cache.set(
            cache_key + fatlink.hash,
            str(error_count),
            75,
        )
    else:
        close_esi_fleet(fatlink=fatlink, reason=logger_message)


def initialize_caches(fatlink: AFatLink) -> None:
    """
    Initializing caches
    :param fatlink:
    :type fatlink:
    :return:
    :rtype:
    """

    if cache.get(CACHE_KEY_FLEET_CHANGED_ERROR + fatlink.hash) is None:
        cache.set(CACHE_KEY_FLEET_CHANGED_ERROR + fatlink.hash, "0", 75)

    if cache.get(CACHE_KEY_NO_FLEETBOSS_ERROR + fatlink.hash) is None:
        cache.set(CACHE_KEY_NO_FLEETBOSS_ERROR + fatlink.hash, "0", 75)

    if cache.get(CACHE_KEY_NO_FLEET_ERROR + fatlink.hash) is None:
        cache.set(CACHE_KEY_NO_FLEET_ERROR + fatlink.hash, "0", 75)

    if cache.get(CACHE_KEY_NOT_IN_FLEET_ERROR + fatlink.hash) is None:
        cache.set(CACHE_KEY_NOT_IN_FLEET_ERROR + fatlink.hash, "0", 75)


@shared_task(**{**TASK_ESI_KWARGS}, **{"base": QueueOnce})
def update_esi_fatlinks() -> None:
    """
    Checking ESI fat links for changes
    :return:
    :rtype:
    """

    required_scopes = ["esi-fleets.read_fleet.v1"]

    try:
        esi_fatlinks = AFatLink.objects.select_related_default().filter(
            is_esilink=True, is_registered_on_esi=True
        )

        for fatlink in esi_fatlinks:
            initialize_caches(fatlink=fatlink)

            logger.info(f'Processing ESI FAT link with hash "{fatlink.hash}"')

            if fatlink.creator.profile.main_character is not None:
                # Check if there is a fleet
                try:
                    fleet_commander_id = fatlink.character.character_id
                    esi_token = Token.get_token(fleet_commander_id, required_scopes)

                    fleet_from_esi = (
                        esi.client.Fleets.get_characters_character_id_fleet(
                            character_id=fleet_commander_id,
                            token=esi_token.valid_access_token(),
                        ).result()
                    )

                    if fatlink.esi_fleet_id == fleet_from_esi["fleet_id"]:
                        # Check if we deal with the fleet boss here
                        try:
                            esi_fleet_member = (
                                esi.client.Fleets.get_fleets_fleet_id_members(
                                    fleet_id=fleet_from_esi["fleet_id"],
                                    token=esi_token.valid_access_token(),
                                ).result()
                            )

                            # Process fleet members
                            process_fats.delay(
                                data_list=esi_fleet_member,
                                data_source="esi",
                                fatlink_hash=fatlink.hash,
                            )
                        except Exception:
                            esi_fatlinks_error_handling(
                                cache_key=CACHE_KEY_NO_FLEETBOSS_ERROR,
                                fatlink=fatlink,
                                logger_message="FC is no longer the fleet boss.",
                            )

                    else:
                        esi_fatlinks_error_handling(
                            cache_key=CACHE_KEY_FLEET_CHANGED_ERROR,
                            fatlink=fatlink,
                            logger_message="FC switched to another fleet",
                        )

                except HTTPNotFound:
                    esi_fatlinks_error_handling(
                        cache_key=CACHE_KEY_NOT_IN_FLEET_ERROR,
                        fatlink=fatlink,
                        logger_message=(
                            "FC is not in the registered fleet anymore or "
                            "fleet is no longer available."
                        ),
                    )

                except Exception:
                    esi_fatlinks_error_handling(
                        cache_key=CACHE_KEY_NO_FLEET_ERROR,
                        fatlink=fatlink,
                        logger_message="Registered fleet is no longer available.",
                    )

            else:
                close_esi_fleet(fatlink=fatlink, reason="No fatlink creator available.")

    except AFatLink.DoesNotExist:
        pass


@shared_task
def logrotate():
    """
    Remove logs older than AFAT_DEFAULT_LOG_DURATION
    :return:
    :rtype:
    """

    logger.info(f"Cleaning up logs older than {AFAT_DEFAULT_LOG_DURATION} days")

    AFatLog.objects.filter(
        log_time__lte=timezone.now() - timedelta(days=AFAT_DEFAULT_LOG_DURATION)
    ).delete()



"""
Fat links related views
"""

from datetime import datetime, timedelta

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest

# from django.db.models import Count, F
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from allianceauth.authentication.decorators import permissions_required
from allianceauth.eveonline.models import EveCharacter
from allianceauth.eveonline.providers import provider
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from esi.decorators import token_required
from esi.models import Token

from afat import __title__
from afat.app_settings import (
    AFAT_DEFAULT_FATLINK_EXPIRY_TIME,
    AFAT_DEFAULT_FATLINK_REOPEN_DURATION,
    AFAT_DEFAULT_FATLINK_REOPEN_GRACE_TIME,
)
from afat.forms import (  # ExtendFatLinkDuration,
    AFatClickFatForm,
    AFatEsiFatForm,
    AFatManualFatForm,
    FatLinkEditForm,
)
from afat.helper.fatlinks import get_esi_fleet_information_by_user
from afat.helper.time import get_time_delta
from afat.helper.views_helper import convert_fatlinks_to_dict, convert_fats_to_dict
from afat.models import (
    AFat,
    AFatLink,
    AFatLinkType,
    AFatLogEvent,
    ClickAFatDuration,
    get_hash_on_save,
)
from afat.providers import esi
from afat.tasks import process_fats
from afat.utils import get_or_create_character, write_log

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required()
@permission_required("afat.basic_access")
def overview(request: WSGIRequest, year: int = None) -> HttpResponse:
    """
    Fat links view
    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    if year is None:
        year = datetime.now().year

    context = {
        "year": year,
        "year_current": datetime.now().year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
    }

    logger.info(f"FAT link list called by {request.user}")

    return render(request, "afat/view/fatlinks/fatlinks_overview.html", context)


@login_required()
@permission_required("afat.basic_access")
def ajax_get_fatlinks_by_year(request: WSGIRequest, year: int) -> JsonResponse:
    """
    Ajax call :: get all FAT links for a given year
    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    fatlinks = (
        AFatLink.objects.select_related_default()
        .filter(afattime__year=year)
        .annotate_afats_count()
    )

    fatlink_rows = [
        convert_fatlinks_to_dict(
            request=request,
            fatlink=fatlink,
            close_esi_redirect=reverse("afat:fatlinks_overview"),
        )
        for fatlink in fatlinks
    ]

    return JsonResponse(fatlink_rows, safe=False)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def add_fatlink(request: WSGIRequest) -> HttpResponse:
    """
    Add fat link view
    :param request:
    :type request:
    :return:
    :rtype:
    """

    link_types_configured = False
    link_types_count = AFatLinkType.objects.all().count()

    if link_types_count > 0:
        link_types_configured = True

    context = {
        "link_types_configured": link_types_configured,
        "default_expiry_time": AFAT_DEFAULT_FATLINK_EXPIRY_TIME,
        "esi_fleet": get_esi_fleet_information_by_user(request.user),
        "esi_fatlink_form": AFatEsiFatForm(),
        "manual_fatlink_form": AFatClickFatForm(),
    }

    logger.info(f"Add FAT link view called by {request.user}")

    return render(request, "afat/view/fatlinks/fatlinks_add_fatlink.html", context)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def create_clickable_fatlink(
    request: WSGIRequest,
) -> HttpResponseRedirect:
    """
    Create clickable fat link
    :param request:
    :type request:
    :return:
    :rtype:
    """

    if request.method == "POST":
        form = AFatClickFatForm(request.POST)

        if form.is_valid():
            fatlink_hash = get_hash_on_save()
            # fatlink_hash = get_random_string(length=30)

            fatlink = AFatLink()
            fatlink.fleet = form.cleaned_data["name"]

            if form.cleaned_data["type"] is not None:
                fatlink.link_type = form.cleaned_data["type"]

            fatlink.creator = request.user
            fatlink.hash = fatlink_hash
            fatlink.afattime = timezone.now()
            fatlink.save()

            dur = ClickAFatDuration()
            dur.fleet = AFatLink.objects.get(hash=fatlink_hash)
            dur.duration = form.cleaned_data["duration"]
            dur.save()

            # Writing DB log
            fleet_type = ""
            if fatlink.link_type:
                fleet_type = f" (Fleet Type: {fatlink.link_type.name})"

            write_log(
                request=request,
                log_event=AFatLogEvent.CREATE_FATLINK,
                log_text=(
                    f'FAT link with name "{form.cleaned_data["name"]}"{fleet_type} and '
                    f'a duration of {form.cleaned_data["duration"]} minutes was created'
                ),
                fatlink_hash=fatlink.hash,
            )

            logger.info(
                f'FAT link "{fatlink_hash}" with name '
                f'"{form.cleaned_data["name"]}"{fleet_type} and a duration '
                f'of {form.cleaned_data["duration"]} minutes was created '
                f"by {request.user}"
            )

            messages.success(
                request,
                mark_safe(
                    _(
                        "<h4>Success!</h4>"
                        "<p>Clickable FAT Link Created!</p>"
                        "<p>Make sure to give your fleet members the link to "
                        "click so that they get credit for this fleet.</p>"
                    )
                ),
            )

            return redirect("afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)

        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>Something went wrong when attempting "
                    "to submit your clickable FAT Link.</p>"
                )
            ),
        )

        return redirect("afat:fatlinks_add_fatlink")

    messages.warning(
        request,
        mark_safe(
            _(
                "<h4>Warning!</h4>"
                '<p>You must fill out the form on the "Add FAT Link" '
                "page to create a clickable FAT Link</p>"
            )
        ),
    )

    return redirect("afat:fatlinks_add_fatlink")


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
@token_required(scopes=["esi-fleets.read_fleet.v1"])
def create_esi_fatlink_callback(
    request: WSGIRequest, token, fatlink_hash: str
) -> HttpResponseRedirect:
    """
    Helper :: create ESI link (callback, used when coming back from character selection)
    :param request:
    :type request:
    :param token:
    :type token:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    # Check if there is a fleet
    try:
        required_scopes = ["esi-fleets.read_fleet.v1"]
        esi_token = Token.get_token(token.character_id, required_scopes)

        fleet_from_esi = esi.client.Fleets.get_characters_character_id_fleet(
            character_id=token.character_id, token=esi_token.valid_access_token()
        ).result()
    except Exception:
        # Not in a fleet
        messages.warning(
            request,
            mark_safe(
                _(
                    "<h4>Warning!</h4>"
                    "<p>To use the ESI function, you need to be in fleet and you need "
                    "to be the fleet boss! You can create a clickable FAT link and "
                    "share it, if you like.</p>"
                )
            ),
        )

        # Return to "Add FAT Link" view
        return redirect("afat:fatlinks_add_fatlink")

    # check if this character already has a fleet
    creator_character = EveCharacter.objects.get(character_id=token.character_id)
    registered_fleets_for_creator = AFatLink.objects.select_related_default().filter(
        is_esilink=True,
        is_registered_on_esi=True,
        character__character_name=creator_character.character_name,
    )

    fleet_already_registered = False
    character_has_registered_fleets = False
    registered_fleets_to_close = list()

    if registered_fleets_for_creator.count() > 0:
        character_has_registered_fleets = True

        for registered_fleet in registered_fleets_for_creator:
            if registered_fleet.esi_fleet_id == fleet_from_esi["fleet_id"]:
                # Character already has a fleet
                fleet_already_registered = True
            else:
                registered_fleets_to_close.append(
                    {"registered_fleet": registered_fleet}
                )

    # If the FC already has a fleet, and it is the same as already registered,
    # just throw a warning
    if fleet_already_registered is True:
        messages.warning(
            request,
            mark_safe(
                _(
                    "<h4>Warning!</h4>"
                    f'<p>Fleet with ID "{fleet_from_esi["fleet_id"]}" for your '
                    f"character {creator_character.character_name} has already been "
                    f"registered and pilots joining this fleet are automatically "
                    f"tracked.</p>"
                )
            ),
        )

        # Return to "Add FAT Link" view
        return redirect("afat:fatlinks_add_fatlink")

    # If it's a new fleet, remove all former registered fleets, if there are any
    if (
        character_has_registered_fleets is True
        and fleet_already_registered is False
        and len(registered_fleets_to_close) > 0
    ):
        for registered_fleet_to_close in registered_fleets_to_close:
            reason = (
                f"FC has opened a new fleet with the "
                f"character {creator_character.character_name}"
            )

            logger.info(
                f"Closing ESI FAT link with hash "
                f'"{registered_fleet_to_close["registered_fleet"].hash}". '
                f"Reason: {reason}"
            )

            registered_fleet_to_close["registered_fleet"].is_registered_on_esi = False
            registered_fleet_to_close["registered_fleet"].save()

    # Check if we deal with the fleet boss here
    try:
        esi_fleet_member = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=fleet_from_esi["fleet_id"],
            token=esi_token.valid_access_token(),
        ).result()
    except Exception:
        messages.warning(
            request,
            mark_safe(
                _(
                    "<h4>Warning!</h4>"
                    "<p>Not Fleet Boss! Only the fleet boss can utilize the ESI "
                    "function. You can create a clickable FAT link and share it, "
                    "if you like.</p>"
                )
            ),
        )

        # Return to "Add FAT Link" view
        return redirect("afat:fatlinks_add_fatlink")

    creator_character = EveCharacter.objects.get(character_id=token.character_id)

    # Create the fat link
    fatlink = AFatLink(
        afattime=timezone.now(),
        fleet=request.session["fatlink_form__name"],
        creator=request.user,
        character=creator_character,
        hash=fatlink_hash,
        is_esilink=True,
        is_registered_on_esi=True,
        esi_fleet_id=fleet_from_esi["fleet_id"],
    )

    # Add fleet type, if there is any
    if request.session["fatlink_form__type"] is not None:
        fatlink.link_type_id = request.session["fatlink_form__type"]

    # Save it
    fatlink.save()

    # Writing DB log
    fleet_type = ""
    if fatlink.link_type:
        fleet_type = f" (Fleet Type: {fatlink.link_type.name})"

    write_log(
        request=request,
        log_event=AFatLogEvent.CREATE_FATLINK,
        log_text=(
            f'ESI FAT link with name "{request.session["fatlink_form__name"]}"'
            f"{fleet_type} was created by {request.user}"
        ),
        fatlink_hash=fatlink.hash,
    )

    logger.info(
        f'ESI FAT link "{fatlink_hash}" with name '
        f'"{request.session["fatlink_form__name"]}"{fleet_type} '
        f"was created by {request.user}"
    )

    # Clear session
    del request.session["fatlink_form__name"]
    del request.session["fatlink_form__type"]

    # Process fleet members in background
    process_fats.delay(
        data_list=esi_fleet_member, data_source="esi", fatlink_hash=fatlink_hash
    )

    messages.success(
        request,
        mark_safe(
            _(
                "<h4>Success!</h4>"
                "<p>FAT Link Created!</p>"
                "<p>FATs have been queued, they may take a few mins to show up.</p>"
                "<p>Pilots who join later will be automatically added until you "
                "close or leave the fleet in-game.</p>"
            )
        ),
    )

    return redirect("afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)


@login_required()
def create_esi_fatlink(
    request: WSGIRequest,
) -> HttpResponseRedirect:
    """
    Create ESI fat link
    :param request:
    :type request:
    :return:
    :rtype:
    """

    fatlink_form = AFatEsiFatForm(request.POST)

    if fatlink_form.is_valid():
        fatlink_hash = get_hash_on_save()
        # fatlink_hash = get_random_string(length=30)

        fatlink_type = None
        if fatlink_form.cleaned_data["type_esi"]:
            fatlink_type_from_form = fatlink_form.cleaned_data["type_esi"]
            fatlink_type = fatlink_type_from_form.pk

        request.session["fatlink_form__name"] = fatlink_form.cleaned_data["name_esi"]
        request.session["fatlink_form__type"] = fatlink_type

        return redirect(
            "afat:fatlinks_create_esi_fatlink_callback", fatlink_hash=fatlink_hash
        )

    messages.error(
        request,
        mark_safe(
            _(
                "<h4>Error!</h4>"
                "<p>Something went wrong when attempting to "
                "submit your ESI FAT Link.</p>"
            )
        ),
    )

    return redirect("afat:fatlinks_add_fatlink")


@login_required()
@permission_required("afat.basic_access")
@token_required(
    scopes=[
        "esi-location.read_location.v1",
        "esi-location.read_ship_type.v1",
        "esi-location.read_online.v1",
    ]
)
def add_fat(
    request: WSGIRequest, token, fatlink_hash: str = None
) -> HttpResponseRedirect:
    """
    Click fat link helper
    :param request:
    :type request:
    :param token:
    :type token:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    if fatlink_hash is None:
        messages.warning(
            request,
            mark_safe(_("<h4>Warning!</h4><p>No FAT link hash provided.</p>")),
        )

        return redirect("afat:dashboard")

    try:
        try:
            fleet = AFatLink.objects.get(hash=fatlink_hash)
        except AFatLink.DoesNotExist:
            messages.warning(
                request,
                mark_safe(_("<h4>Warning!</h4><p>The hash provided is not valid.</p>")),
            )

            return redirect("afat:dashboard")

        dur = ClickAFatDuration.objects.get(fleet=fleet)
        now = timezone.now() - timedelta(minutes=dur.duration)

        if now >= fleet.afattime:
            messages.warning(
                request,
                mark_safe(
                    _(
                        "<h4>Warning!</h4>"
                        "<p>Sorry, that FAT Link is expired. "
                        "If you were on that fleet, contact your FC about "
                        "having your FAT manually added.</p>"
                    )
                ),
            )

            return redirect("afat:dashboard")

        character = EveCharacter.objects.get(character_id=token.character_id)

        try:
            required_scopes = [
                "esi-location.read_location.v1",
                "esi-location.read_online.v1",
                "esi-location.read_ship_type.v1",
            ]
            esi_token = Token.get_token(token.character_id, required_scopes)

            # Check if character is online
            character_online = esi.client.Location.get_characters_character_id_online(
                character_id=token.character_id, token=esi_token.valid_access_token()
            ).result()

            if character_online["online"] is True:
                # Character location
                location = esi.client.Location.get_characters_character_id_location(
                    character_id=token.character_id,
                    token=esi_token.valid_access_token(),
                ).result()

                # Current ship
                ship = esi.client.Location.get_characters_character_id_ship(
                    character_id=token.character_id,
                    token=esi_token.valid_access_token(),
                ).result()

                # System information
                system = esi.client.Universe.get_universe_systems_system_id(
                    system_id=location["solar_system_id"]
                ).result()["name"]

                ship_name = provider.get_itemtype(ship["ship_type_id"]).name

                try:
                    fat = AFat(
                        afatlink=fleet,
                        character=character,
                        system=system,
                        shiptype=ship_name,
                    )
                    fat.save()

                    if fleet.fleet is not None:
                        fleet_name = fleet.fleet
                    else:
                        fleet_name = fleet.hash

                    messages.success(
                        request,
                        mark_safe(
                            _(
                                "<h4>Success!</h4>"
                                f"<p>FAT registered for {character.character_name} "
                                f"at {fleet_name}</p>"
                            )
                        ),
                    )

                    logger.info(
                        f'Participation for fleet "{fleet_name}" registered '
                        f"for pilot {character.character_name}"
                    )

                    return redirect("afat:dashboard")
                except Exception:
                    messages.warning(
                        request,
                        mark_safe(
                            _(
                                "<h4>Warning!</h4>"
                                f"<p>A FAT already exists for the selected "
                                f"character ({character.character_name}) and fleet "
                                f"combination.</p>"
                            )
                        ),
                    )

                    return redirect("afat:dashboard")
            else:
                messages.warning(
                    request,
                    mark_safe(
                        _(
                            "<h4>Warning!</h4>"
                            f"<p>Cannot register the fleet participation "
                            f"for {character.character_name}. "
                            f"The character needs to be online.</p>"
                        )
                    ),
                )

                return redirect("afat:dashboard")
        except Exception:
            messages.warning(
                request,
                mark_safe(
                    _(
                        "<h4>Warning!</h4>"
                        f"<p>There was an issue with the ESI token "
                        f"for {character.character_name}. Please try again.</p>"
                    )
                ),
            )

            return redirect("afat:dashboard")
    except Exception:
        messages.warning(
            request,
            mark_safe(
                _(
                    "<h4>Warning!</h4>"
                    "<p>The hash provided is not for a clickable FAT Link.</p>"
                )
            ),
        )

        return redirect("afat:dashboard")


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def details_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponse:
    """
    Fat link view
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        link = AFatLink.objects.select_related_default().get(hash=fatlink_hash)
    except AFatLink.DoesNotExist:
        messages.warning(
            request,
            mark_safe(_("<h4>Warning!</h4><p>The hash provided is not valid.</p>")),
        )

        return redirect("afat:dashboard")

    if request.method == "POST":
        fatlink_edit_form = FatLinkEditForm(request.POST)
        manual_fat_form = AFatManualFatForm(request.POST)

        if fatlink_edit_form.is_valid():
            link.fleet = fatlink_edit_form.cleaned_data["fleet"]
            link.save()

            # Writing DB log
            write_log(
                request=request,
                log_event=AFatLogEvent.CHANGE_FATLINK,
                log_text=f'FAT link changed. Fleet name was set to "{link.fleet}"',
                fatlink_hash=link.hash,
            )

            logger.info(
                f'FAT link with hash "{link.hash}" changed. '
                f'Fleet name was set to "{link.fleet}" by {request.user}'
            )

            messages.success(
                request,
                mark_safe(
                    _("<h4>Success!</h4><p>Fleet name successfully changed.</p>")
                ),
            )
        elif manual_fat_form.is_valid():
            character_name = manual_fat_form.cleaned_data["character"]
            system = manual_fat_form.cleaned_data["system"]
            shiptype = manual_fat_form.cleaned_data["shiptype"]
            character = get_or_create_character(name=character_name)

            if character is not None:
                AFat(
                    afatlink_id=link.pk,
                    character=character,
                    system=system,
                    shiptype=shiptype,
                ).save()

                messages.success(
                    request,
                    mark_safe(_("<h4>Success!</h4><p>Manual FAT processed.</p>")),
                )

                # Writing DB log
                write_log(
                    request=request,
                    log_event=AFatLogEvent.MANUAL_FAT,
                    log_text=(
                        f"Pilot {character.character_name} "
                        f"flying a {shiptype} was manually added"
                    ),
                    fatlink_hash=link.hash,
                )

                logger.info(
                    f"Pilot {character.character_name} flying a {shiptype} was"
                    f" manually added to FAT link with "
                    f'hash "{link.hash}" by {request.user}'
                )
            else:
                messages.error(
                    request,
                    mark_safe(
                        _(
                            "<h4>Oh No!</h4>"
                            "<p>Manual FAT processing failed! "
                            "The character name you entered was not found.</p>"
                        )
                    ),
                )
        else:
            messages.error(
                request,
                mark_safe(_("<h4>Oh No!</h4><p>Something went wrong!</p>")),
            )

    logger.info(f'FAT link "{fatlink_hash}" details view called by {request.user}')

    # Let's see if the link is still valid or has expired already and can be re-opened
    # and FATs can be manually added
    # (only possible for 24 hours after creating the FAT link)
    link_ongoing = True
    link_can_be_reopened = False
    link_expires = None
    manual_fat_can_be_added = False

    # Time dependant settings
    try:
        dur = ClickAFatDuration.objects.get(fleet=link)
        link_expires = link.afattime + timedelta(minutes=dur.duration)
        now = timezone.now()

        if link_expires <= now:
            # Link expired
            link_ongoing = False

            if (
                link.reopened is False
                and get_time_delta(link_expires, now, "minutes")
                < AFAT_DEFAULT_FATLINK_REOPEN_GRACE_TIME
            ):
                link_can_be_reopened = True

        # Manual fat still possible?
        # Only possible if the FAT link has not been re-opened
        # and has been created within the last 24 hours
        if link.reopened is False and get_time_delta(link.afattime, now, "hours") < 24:
            manual_fat_can_be_added = True
    except ClickAFatDuration.DoesNotExist:
        # ESI link
        link_ongoing = False

    is_clickable_link = False
    if link.is_esilink is False:
        is_clickable_link = True

    if link.is_esilink and link.is_registered_on_esi:
        link_ongoing = True

    context = {
        "link": link,
        "is_esi_link": link.is_esilink,
        "is_clickable_link": is_clickable_link,
        "link_expires": link_expires,
        "link_ongoing": link_ongoing,
        "link_can_be_reopened": link_can_be_reopened,
        "manual_fat_can_be_added": manual_fat_can_be_added,
        "reopen_grace_time": AFAT_DEFAULT_FATLINK_REOPEN_GRACE_TIME,
        "reopen_duration": AFAT_DEFAULT_FATLINK_REOPEN_DURATION,
    }

    return render(request, "afat/view/fatlinks/fatlinks_details_fatlink.html", context)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def ajax_get_fats_by_fatlink(request: WSGIRequest, fatlink_hash) -> JsonResponse:
    """
    Ajax call :: get all FATs for a given FAT link hash
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    fats = AFat.objects.select_related_default().filter(afatlink__hash=fatlink_hash)

    fat_rows = [convert_fats_to_dict(request=request, fat=fat) for fat in fats]

    return JsonResponse(fat_rows, safe=False)


@login_required()
@permission_required("afat.manage_afat")
def delete_fatlink(
    request: WSGIRequest, fatlink_hash: str = None
) -> HttpResponseRedirect:
    """
    Delete fat link helper
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    if fatlink_hash is None:
        messages.warning(
            request,
            mark_safe(_("<h4>Warning!</h4><p>No FAT Link hash provided.</p>")),
        )

        return redirect("afat:dashboard")

    try:
        link = AFatLink.objects.get(hash=fatlink_hash)
    except AFatLink.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>The fatlink hash provided is either invalid "
                    "or the fatlink has already been deleted.</p>"
                )
            ),
        )

        return redirect("afat:dashboard")

    AFat.objects.filter(afatlink_id=link.pk).delete()

    link.delete()

    write_log(
        request=request,
        log_event=AFatLogEvent.DELETE_FATLINK,
        log_text="FAT link deleted.",
        fatlink_hash=link.hash,
    )

    messages.success(
        request,
        mark_safe(
            _(
                "<h4>Success!</h4>"
                f'<p>The FAT Link "{fatlink_hash}" and all associated FATs '
                f"have been successfully deleted.</p>"
            )
        ),
    )

    logger.info(
        f'Fat link "{fatlink_hash}" and all associated '
        f"FATs have been deleted by {request.user}"
    )

    return redirect("afat:fatlinks_overview")


@login_required()
@permissions_required(("afat.manage_afat", "afat.delete_afat"))
def delete_fat(request: WSGIRequest, fatlink_hash: str, fat) -> HttpResponseRedirect:
    """
    Delete fat helper
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :param fat:
    :type fat:
    :return:
    :rtype:
    """

    try:
        link = AFatLink.objects.get(hash=fatlink_hash)
    except AFatLink.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>The hash provided is either invalid or has been deleted.</p>"
                )
            ),
        )

        return redirect("afat:dashboard")

    try:
        fat_details = AFat.objects.get(pk=fat, afatlink_id=link.pk)
    except AFat.DoesNotExist:
        messages.error(
            request,
            mark_safe(_("<h4>Error!</h4><p>The hash and FAT ID do not match.</p>")),
        )

        return redirect("afat:dashboard")

    fat_details.delete()

    write_log(
        request=request,
        log_event=AFatLogEvent.DELETE_FAT,
        log_text=f"The FAT for {fat_details.character.character_name} has been deleted",
        fatlink_hash=link.hash,
    )

    messages.success(
        request,
        mark_safe(
            _(
                "<h4>Success!</h4>"
                f"<p>The FAT for {fat_details.character.character_name} "
                f'has been successfully deleted from FAT link "{fatlink_hash}".</p>'
            )
        ),
    )

    logger.info(
        f"The FAT for {fat_details.character.character_name} has "
        f'been deleted from FAT link "{fatlink_hash}" by {request.user}.'
    )

    return redirect("afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def close_esi_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponseRedirect:
    """
    Ajax call to close an ESI fat link
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        fatlink = AFatLink.objects.get(hash=fatlink_hash)

        logger.info(
            f'Closing ESI FAT link with hash "{fatlink_hash}". '
            f"Reason: Closed by manual request"
        )

        fatlink.is_registered_on_esi = False
        fatlink.save()
    except AFatLink.DoesNotExist:
        logger.info(f'ESI FAT link with hash "{fatlink_hash}" does not exist')

    default_redirect = reverse("afat:dashboard")
    next_view = request.GET.get("next", default_redirect)

    return HttpResponseRedirect(next_view)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def reopen_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponseRedirect:
    """
    Re-open fat link
    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    # if request.method == "POST":
    #     fatlink_reopen_form = ExtendFatLinkDuration(request.POST)
    #
    #     if fatlink_reopen_form.is_valid():
    #         duration = ClickAFatDuration.objects.get(fleet__hash=fatlink_hash)
    #         reopen_for = fatlink_reopen_form.cleaned_data["duration"]
    #
    #         # get minutes already passed since fatlink creation
    #         created_at = duration.fleet.afattime
    #         now = datetime.now()
    #
    #         time_difference_in_minutes = get_time_delta(created_at, now, "minutes")
    #         new_duration = (
    #             time_difference_in_minutes
    #             + fatlink_reopen_form.cleaned_data["duration"]
    #         )
    #
    #         duration.duration = new_duration
    #         duration.save()

    try:
        fatlink_duration = ClickAFatDuration.objects.get(fleet__hash=fatlink_hash)

        if fatlink_duration.fleet.reopened is False:
            created_at = fatlink_duration.fleet.afattime
            now = datetime.now()

            time_difference_in_minutes = get_time_delta(created_at, now, "minutes")
            new_duration = (
                time_difference_in_minutes + AFAT_DEFAULT_FATLINK_REOPEN_DURATION
            )

            fatlink_duration.duration = new_duration
            fatlink_duration.save()

            fatlink_duration.fleet.reopened = True
            fatlink_duration.fleet.save()

            # writing DB log
            write_log(
                request=request,
                # log_event=AFatLogEvent.REOPEN_FATLINK,
                log_event=AFatLogEvent.REOPEN_FATLINK,
                log_text=(
                    f"FAT link re-opened for a "
                    f"duration of {AFAT_DEFAULT_FATLINK_REOPEN_DURATION} minutes"
                ),
                fatlink_hash=fatlink_duration.fleet.hash,
            )

            logger.info(
                f'FAT link with hash "{fatlink_hash}" '
                f"re-opened by {request.user} for a "
                f"duration of {AFAT_DEFAULT_FATLINK_REOPEN_DURATION} minutes"
            )

            messages.success(
                request,
                mark_safe(
                    _(
                        "<h4>Success!</h4>"
                        "<p>The FAT link has been successfully re-opened.</p>"
                    )
                ),
            )
        else:
            messages.warning(
                request,
                mark_safe(
                    _(
                        "<h4>Warning!</h4>"
                        "<p>This FAT link has already been re-opened. "
                        "FAT links can be re-opened only once!</p>"
                    )
                ),
            )
    except ClickAFatDuration.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>The hash you provided does not match with any FAT link.</p>"
                )
            ),
        )

    return redirect("afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)



"""
Import FAT data from alliance auth fat module
"""

from django.apps import apps
from django.core.management.base import BaseCommand

from allianceauth.fleetactivitytracking.models import Fat, Fatlink

from afat.models import AFat, AFatLink, AFatLog, AFatLogEvent


def get_input(text) -> str:
    """
    Wrapped input to enable import
    :param text:
    :type text:
    :return:
    :rtype:
    """

    return input(text)


def aa_fat_installed() -> bool:
    """
    Check if native fat is installed
    :return:
    :rtype:
    """

    return apps.is_installed("allianceauth.fleetactivitytracking")


class Command(BaseCommand):
    """
    Initial import of FAT data from AA FAT module
    """

    help = "Imports FAT data from AA FAT module"

    def _import_from_aa_fat(self) -> None:
        """
        Start the import
        :return:
        :rtype:
        """

        # Check if AA FAT is active
        if aa_fat_installed():
            self.stdout.write(
                self.style.SUCCESS("Alliance Auth FAT module is active, let's go!")
            )

            # First, we check if the target tables are empty ...
            current_afat_links_count = AFatLink.objects.all().count()
            current_afat_count = AFat.objects.all().count()

            if current_afat_count > 0 or current_afat_links_count > 0:
                self.stdout.write(
                    self.style.WARNING(
                        "You already have FAT data with the AFAT module. "
                        "Import cannot be continued."
                    )
                )

                return

            aa_fatlinks = Fatlink.objects.all()
            for aa_fatlink in aa_fatlinks:
                self.stdout.write(
                    "Importing FAT link for fleet '{fleet}' with hash "
                    "'{fatlink_hash}'.".format(
                        fleet=aa_fatlink.fleet, fatlink_hash=aa_fatlink.hash
                    )
                )

                afatlink = AFatLink()

                afatlink.id = aa_fatlink.id
                afatlink.afattime = aa_fatlink.fatdatetime
                afatlink.fleet = (
                    aa_fatlink.fleet
                    if aa_fatlink.fleet is not None
                    else aa_fatlink.hash
                )
                afatlink.hash = aa_fatlink.hash
                afatlink.creator_id = aa_fatlink.creator_id

                afatlink.save()

                # Write to log table
                log_text = (
                    "FAT link {fatlink_hash} with name {name} was created by {user}"
                ).format(
                    fatlink_hash=aa_fatlink.hash,
                    name=aa_fatlink.fleet,
                    user=aa_fatlink.creator,
                )

                afatlog = AFatLog()
                afatlog.log_time = aa_fatlink.fatdatetime
                afatlog.log_event = AFatLogEvent.CREATE_FATLINK
                afatlog.log_text = log_text
                afatlog.user_id = aa_fatlink.creator_id
                afatlog.save()

            aa_fats = Fat.objects.all()
            for aa_fat in aa_fats:
                self.stdout.write(
                    "Importing FATs for FAT link ID '{fatlink_id}'.".format(
                        fatlink_id=aa_fat.id
                    )
                )

                afat = AFat()

                afat.id = aa_fat.id
                afat.system = aa_fat.system
                afat.shiptype = aa_fat.shiptype
                afat.character_id = aa_fat.character_id
                afat.afatlink_id = aa_fat.fatlink_id

                afat.save()

            self.stdout.write(
                self.style.SUCCESS(
                    "Import complete! "
                    "You can now deactivate the Alliance Auth FAT "
                    "module in your local.py"
                )
            )
        else:
            self.stdout.write(
                self.style.WARNING(
                    "Alliance Auth FAT module is not active. "
                    "Please make sure you have it in your "
                    "INSTALLED_APPS in your local.py!"
                )
            )

    def handle(self, *args, **options):
        """
        Ask before running ...
        :param args:
        :type args:
        :param options:
        :type options:
        :return:
        :rtype:
        """

        self.stdout.write(
            "Importing all FAT/FAT link data from Alliance Auth's built in FAT module. "
            "This can only be done once during the very first installation. "
            "As soon as you have data collected with your AFAT module, "
            "this import will fail!"
        )

        user_input = get_input("Are you sure you want to proceed? (yes/no)?")

        if user_input == "yes":
            self.stdout.write("Starting import. Please stand by.")
            self._import_from_aa_fat()
        else:
            self.stdout.write(self.style.WARNING("Aborted."))



# Generated by Django 3.1.1 on 2020-09-25 22:06

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import afat.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveonline", "0012_index_additions"),
        ("afat", "0008_auto_20200912_1116"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="afatlinktype",
            options={
                "default_permissions": (),
                "verbose_name": "FAT Link Fleet Type",
                "verbose_name_plural": "FAT Link Fleet Types",
            },
        ),
        migrations.AddField(
            model_name="afatlink",
            name="is_esilink",
            field=models.BooleanField(
                default=False,
                help_text="Whether this fatlink was created via ESI or not",
            ),
        ),
        migrations.AddField(
            model_name="afatlinktype",
            name="is_enabled",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="Whether this fleettype is active or not",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="afatlink",
            field=models.ForeignKey(
                help_text="The fatlink the character registered at",
                on_delete=django.db.models.deletion.CASCADE,
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="character",
            field=models.ForeignKey(
                help_text="Character who registered this fat",
                on_delete=django.db.models.deletion.CASCADE,
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="deleted_at",
            field=models.DateTimeField(
                blank=True, help_text="Has been deleted or not", null=True
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="shiptype",
            field=models.CharField(
                help_text="The ship the character was flying", max_length=100, null=True
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="system",
            field=models.CharField(
                help_text="The system the character is in", max_length=100, null=True
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="afattime",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="When was this fatlink created",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="creator",
            field=models.ForeignKey(
                help_text="Who created the fatlink?",
                on_delete=models.SET(afat.models.get_sentinel_user),
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="deleted_at",
            field=models.DateTimeField(
                blank=True, help_text="Has been deleted or not", null=True
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="fleet",
            field=models.CharField(
                help_text="The fatlinks fleet name", max_length=254, null=True
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="hash",
            field=models.CharField(help_text="The fatlinks hash", max_length=254),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="link_type",
            field=models.ForeignKey(
                help_text="The fatlinks fleet type, if it's set",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="afat.afatlinktype",
            ),
        ),
        migrations.AlterField(
            model_name="afatlinktype",
            name="deleted_at",
            field=models.DateTimeField(
                blank=True, help_text="Has this been deleted?", null=True
            ),
        ),
        migrations.AlterField(
            model_name="afatlinktype",
            name="name",
            field=models.CharField(
                help_text="Descriptive name of your fleet type", max_length=254
            ),
        ),
    ]



# Generated by Django 2.0.6 on 2018-06-26 01:44

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import afat.models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0010_alliance_ticker"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("afat", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ManualAFat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.EveCharacter",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        on_delete=models.SET(afat.models.get_sentinel_user),
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "afatlink",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="afat.AFatLink",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-15 16:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("afat", "0012_manualafat_created_at"),
    ]

    operations = [
        migrations.CreateModel(
            name="AaAfat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "Alliance Auth AFAT",
                "permissions": (
                    ("basic_access", "Can access the Alliance Auth AFAT module"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
    ]



# Generated by Django 2.0.8 on 2018-11-13 10:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("afat", "0003_auto_20180911_0831"),
    ]

    operations = [
        migrations.CreateModel(
            name="ClickAFatDuration",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("duration", models.PositiveIntegerField()),
                (
                    "fleet",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="afat.AFatLink",
                    ),
                ),
            ],
        ),
    ]



"""
App config
"""

from django.apps import AppConfig

from afat import __version__


class AfatConfig(AppConfig):
    """
    General config
    """

    name = "afat"
    label = "afat"
    verbose_name = f"AFAT - Another Fleet Activity Tracker v{__version__}"



"""
Constants used in this module
"""

from django.utils.text import slugify

from afat import __version__

VERBOSE_NAME = "AFAT Fleet Activity Tracking for Alliance Auth"
USER_AGENT = "{verbose_name} v{version} {github_url}".format(
    verbose_name=slugify(VERBOSE_NAME, allow_unicode=True),
    version=__version__,
    github_url="https://github.com/ppfeufer/allianceauth-afat",
)



"""
Template tags
"""

import calendar

from django.template.defaulttags import register


@register.filter
def month_name(value):
    """
    Template tag :: get month name from month number
    example: {{ event.month|month_name }}
    :param value:
    :type value:
    :return:
    :rtype:
    """

    value = int(value)

    return calendar.month_name[value]



"""
Helper functions for fat links view
"""

from typing import Dict, List, Union

from allianceauth.authentication.admin import User

from afat.models import AFatLink


def get_esi_fleet_information_by_user(
    user: User,
) -> Dict[str, Union[bool, List[Dict[int, AFatLink]]]]:
    """
    Get ESI fleet information by a given FC (user)
    :param user:
    :type user:
    :return:
    :rtype:
    """

    has_open_esi_fleets = False
    open_esi_fleets_list = list()
    open_esi_fleets = (
        AFatLink.objects.select_related_default()
        .filter(creator=user, is_esilink=True, is_registered_on_esi=True)
        .order_by("character__character_name")
    )

    if open_esi_fleets.count() > 0:
        has_open_esi_fleets = True

        for open_esi_fleet in open_esi_fleets:
            open_esi_fleets_list.append(open_esi_fleet)

    return {
        "has_open_esi_fleets": has_open_esi_fleets,
        "open_esi_fleets_list": open_esi_fleets_list,
    }



"""
Test afat helpers
"""
from datetime import timedelta

from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ..helper.fatlinks import get_esi_fleet_information_by_user
from ..helper.time import get_time_delta
from ..helper.views_helper import (
    convert_fatlinks_to_dict,
    convert_fats_to_dict,
    convert_logs_to_dict,
)
from ..models import (
    AFat,
    AFatLink,
    AFatLinkType,
    AFatLog,
    AFatLogEvent,
    ClickAFatDuration,
    get_hash_on_save,
)
from ..utils import get_main_character_from_user, write_log
from .fixtures.load_allianceauth import load_allianceauth

MODULE_PATH = "afat.views.fatlinks"


class TestHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.factory = RequestFactory()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_with_add_fatlink, _ = create_user_from_evecharacter(
            cls.character_1001.character_id,
            permissions=["afat.basic_access", "afat.add_fatlink"],
        )

        add_character_to_user(cls.user_with_add_fatlink, cls.character_1101)

        cls.user_with_manage_afat, _ = create_user_from_evecharacter(
            cls.character_1002.character_id,
            permissions=["afat.basic_access", "afat.manage_afat"],
        )

        add_character_to_user(cls.user_with_add_fatlink, cls.character_1003)

    def test_helper_get_esi_fleet_information_by_user(self):
        # given
        fatlink_hash_fleet_1 = get_hash_on_save()
        fatlink_1 = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 1",
            creator=self.user_with_add_fatlink,
            character=self.character_1001,
            hash=fatlink_hash_fleet_1,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id="3726458287",
        )

        fatlink_hash_fleet_2 = get_hash_on_save()
        fatlink_2 = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 2",
            creator=self.user_with_add_fatlink,
            character=self.character_1101,
            hash=fatlink_hash_fleet_2,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id="372645827",
        )

        self.client.force_login(self.user_with_add_fatlink)

        # when
        response = get_esi_fleet_information_by_user(user=self.user_with_add_fatlink)

        # then
        self.assertDictEqual(
            response,
            {
                "has_open_esi_fleets": True,
                "open_esi_fleets_list": [fatlink_1, fatlink_2],
            },
        )

    def test_helper_get_time_delta(self):
        # given
        duration = 1812345
        now = timezone.now()
        expires = timedelta(minutes=duration) + now

        self.client.force_login(self.user_with_add_fatlink)

        # when
        total = get_time_delta(now, expires)
        years = get_time_delta(now, expires, "years")
        days = get_time_delta(now, expires, "days")
        hours = get_time_delta(now, expires, "hours")
        minutes = get_time_delta(now, expires, "minutes")
        seconds = get_time_delta(now, expires, "seconds")

        # then
        self.assertEqual(total, "3 years, 163 days, 13 hours, 45 minutes and 0 seconds")
        self.assertEqual(years, 3)
        self.assertEqual(days, 1258)
        self.assertEqual(hours, 30205)
        self.assertEqual(minutes, 1812345)
        self.assertEqual(seconds, 108740700)

    def test_helper_convert_fatlinks_to_dict(self):
        # given
        self.client.force_login(self.user_with_manage_afat)
        request = self.factory.get(reverse("afat:dashboard"))
        request.user = self.user_with_manage_afat

        fatlink_hash_fleet_1 = get_hash_on_save()
        fatlink_1_created = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 1",
            creator=self.user_with_manage_afat,
            character=self.character_1001,
            hash=fatlink_hash_fleet_1,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id="3726458287",
        )
        AFat.objects.create(
            character=self.character_1101, afatlink=fatlink_1_created, shiptype="Omen"
        )

        fatlink_type_cta = AFatLinkType.objects.create(name="CTA")
        fatlink_hash_fleet_2 = get_hash_on_save()
        fatlink_2_created = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 2",
            creator=self.user_with_add_fatlink,
            character=self.character_1101,
            hash=fatlink_hash_fleet_2,
            link_type=fatlink_type_cta,
        )
        AFat.objects.create(
            character=self.character_1001, afatlink=fatlink_2_created, shiptype="Omen"
        )

        # when
        fatlink_1 = (
            AFatLink.objects.select_related_default()
            .annotate_afats_count()
            .get(hash=fatlink_hash_fleet_1)
        )
        close_esi_tracking_url = reverse(
            "afat:fatlinks_close_esi_fatlink", args=[fatlink_1.hash]
        )
        edit_url_1 = reverse("afat:fatlinks_details_fatlink", args=[fatlink_1.hash])
        delete_url_1 = reverse("afat:fatlinks_delete_fatlink", args=[fatlink_1.hash])

        fatlink_2 = (
            AFatLink.objects.select_related_default()
            .annotate_afats_count()
            .get(hash=fatlink_hash_fleet_2)
        )
        edit_url_2 = reverse("afat:fatlinks_details_fatlink", args=[fatlink_2.hash])
        delete_url_2 = reverse("afat:fatlinks_delete_fatlink", args=[fatlink_2.hash])

        result_1 = convert_fatlinks_to_dict(request=request, fatlink=fatlink_1)
        result_2 = convert_fatlinks_to_dict(request=request, fatlink=fatlink_2)

        # then
        fleet_time_1 = fatlink_1.afattime
        fleet_time_timestamp_1 = fleet_time_1.timestamp()
        creator_main_character_1 = get_main_character_from_user(user=fatlink_1.creator)
        self.assertDictEqual(
            result_1,
            {
                "pk": fatlink_1.pk,
                "fleet_name": (
                    'April Fleet 1<span class="label label-default '
                    "afat-label afat-label-via-esi "
                    'afat-label-active-esi-fleet">via ESI</span>'
                ),
                "creator_name": creator_main_character_1,
                "fleet_type": "",
                "fleet_time": {
                    "time": fleet_time_1,
                    "timestamp": fleet_time_timestamp_1,
                },
                "fats_number": fatlink_1.afats_count,
                "hash": fatlink_1.hash,
                "is_esilink": True,
                "esi_fleet_id": 3726458287,
                "is_registered_on_esi": True,
                "actions": (
                    f'<a class="btn btn-afat-action btn-primary btn-sm" '
                    f'style="margin-left: 0.25rem;" title="Clicking here will '
                    f"stop the automatic tracking through ESI for this fleet "
                    f'and close the associated FAT link." data-toggle="modal" '
                    f'data-target="#cancelEsiFleetModal" '
                    f'data-url="{close_esi_tracking_url}" '
                    f'data-body-text="<p>Are you sure you want to close ESI '
                    f'fleet with ID 3726458287 from Bruce Wayne?</p>" '
                    f'data-confirm-text="Stop Tracking"><i class="fas '
                    f'fa-times"></i></a><a class="btn btn-afat-action btn-info '
                    f'btn-sm" href="{edit_url_1}"><span class="fas '
                    f'fa-eye"></span></a><a class="btn btn-afat-action '
                    f'btn-danger btn-sm" data-toggle="modal" '
                    f'data-target="#deleteFatLinkModal" '
                    f'data-url="{delete_url_1}" '
                    f'data-confirm-text="Delete"data-body-text="<p>Are you '
                    f"sure you want to delete FAT link April Fleet "
                    f'1?</p>"><span class="glyphicon '
                    f'glyphicon-trash"></span></a>'
                ),
                "via_esi": "Yes",
            },
        )

        fleet_time_2 = fatlink_2.afattime
        fleet_time_timestamp_2 = fleet_time_2.timestamp()
        creator_main_character_2 = get_main_character_from_user(user=fatlink_2.creator)
        self.assertDictEqual(
            result_2,
            {
                "pk": fatlink_2.pk,
                "fleet_name": "April Fleet 2",
                "creator_name": creator_main_character_2,
                "fleet_type": "CTA",
                "fleet_time": {
                    "time": fleet_time_2,
                    "timestamp": fleet_time_timestamp_2,
                },
                "fats_number": fatlink_2.afats_count,
                "hash": fatlink_2.hash,
                "is_esilink": False,
                "esi_fleet_id": None,
                "is_registered_on_esi": False,
                "actions": (
                    f'<a class="btn btn-afat-action btn-info btn-sm" '
                    f'href="{edit_url_2}"><span class="fas '
                    f'fa-eye"></span></a><a class="btn btn-afat-action '
                    f'btn-danger btn-sm" data-toggle="modal" '
                    f'data-target="#deleteFatLinkModal" '
                    f'data-url="{delete_url_2}" '
                    f'data-confirm-text="Delete"data-body-text="<p>Are you '
                    f"sure you want to delete FAT link April Fleet "
                    f'2?</p>"><span class="glyphicon '
                    f'glyphicon-trash"></span></a>'
                ),
                "via_esi": "No",
            },
        )

    def test_helper_convert_fats_to_dict(self):
        # given
        self.client.force_login(self.user_with_manage_afat)
        request = self.factory.get(reverse("afat:dashboard"))
        request.user = self.user_with_manage_afat

        fatlink_hash = get_hash_on_save()
        fatlink_type_cta = AFatLinkType.objects.create(name="CTA")
        fatlink_created = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 1",
            creator=self.user_with_manage_afat,
            character=self.character_1001,
            hash=fatlink_hash,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id="3726458287",
            link_type=fatlink_type_cta,
        )
        fat = AFat.objects.create(
            character=self.character_1101, afatlink=fatlink_created, shiptype="Omen"
        )

        # when
        result = convert_fats_to_dict(request=request, fat=fat)

        esi_marker = (
            '<span class="label label-default afat-label afat-label-via-esi '
            'afat-label-active-esi-fleet">via ESI</span>'
        )
        fleet_time = fat.afatlink.afattime
        fleet_time_timestamp = fleet_time.timestamp()

        button_delete_fat = reverse(
            "afat:fatlinks_delete_fat", args=[fat.afatlink.hash, fat.id]
        )
        button_delete_text = "Delete"
        modal_body_text = (
            "<p>Are you sure you want to remove "
            f"{fat.character.character_name} from this FAT link?</p>"
        )

        # then
        self.assertDictEqual(
            result,
            {
                "system": fat.system,
                "ship_type": fat.shiptype,
                "character_name": fat.character.character_name,
                "fleet_name": fat.afatlink.fleet + esi_marker,
                "fleet_time": {"time": fleet_time, "timestamp": fleet_time_timestamp},
                "fleet_type": "CTA",
                "via_esi": "Yes",
                "actions": (
                    '<a class="btn btn-danger btn-sm" '
                    'data-toggle="modal" '
                    'data-target="#deleteFatModal" '
                    f'data-url="{button_delete_fat}" '
                    f'data-confirm-text="{button_delete_text}"'
                    f'data-body-text="{modal_body_text}">'
                    '<span class="glyphicon glyphicon-trash"></span>'
                    "</a>"
                ),
            },
        )

    def test_helper_convert_logs_to_dict(self):
        # given
        self.client.force_login(self.user_with_manage_afat)
        request = self.factory.get(reverse("afat:dashboard"))
        request.user = self.user_with_manage_afat

        fatlink_hash = get_hash_on_save()
        fatlink_type_cta = AFatLinkType.objects.create(name="CTA")
        fatlink_created = AFatLink.objects.create(
            afattime=timezone.now(),
            fleet="April Fleet 1",
            creator=self.user_with_manage_afat,
            character=self.character_1001,
            hash=fatlink_hash,
            is_esilink=True,
            is_registered_on_esi=True,
            esi_fleet_id="3726458287",
            link_type=fatlink_type_cta,
        )

        duration = ClickAFatDuration.objects.create(fleet=fatlink_created, duration=120)

        fleet_type = f" (Fleet Type: {fatlink_created.link_type.name})"

        write_log(
            request=request,
            log_event=AFatLogEvent.CREATE_FATLINK,
            log_text=(
                f'FAT link with name "{fatlink_created.fleet}"{fleet_type} and '
                f"a duration of {duration.duration} minutes was created"
            ),
            fatlink_hash=fatlink_created.hash,
        )

        # when
        log = AFatLog.objects.get(fatlink_hash=fatlink_hash)
        log_time = log.log_time
        log_time_timestamp = log_time.timestamp()
        user_main_character = get_main_character_from_user(user=log.user)
        fatlink_link = reverse("afat:fatlinks_details_fatlink", args=[log.fatlink_hash])
        fatlink_html = f'<a href="{fatlink_link}">{log.fatlink_hash}</a>'

        result = convert_logs_to_dict(log=log, fatlink_exists=True)

        # then
        self.assertDictEqual(
            result,
            {
                "log_time": {"time": log_time, "timestamp": log_time_timestamp},
                "log_event": AFatLogEvent(log.log_event).label,
                "user": user_main_character,
                "fatlink": {"html": fatlink_html, "hash": log.fatlink_hash},
                "description": log.log_text,
            },
        )
